# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test operations on control flow for dynamic QuantumCircuits."""

import math

from ddt import ddt, data, unpack, idata

from qiskit.circuit import Clbit, ClassicalRegister, Instruction, Parameter, QuantumCircuit, Qubit
from qiskit.circuit.classical import expr, types
from qiskit.circuit.controlflow import CASE_DEFAULT, condition_resources, node_resources
from qiskit.circuit.library import XGate, RXGate
from qiskit.circuit.exceptions import CircuitError

from qiskit.circuit.controlflow import (
    ControlFlowOp,
    WhileLoopOp,
    ForLoopOp,
    IfElseOp,
    ContinueLoopOp,
    BreakLoopOp,
    SwitchCaseOp,
)
from test import QiskitTestCase  # pylint: disable=wrong-import-order


CONDITION_PARAMETRISATION = (
    (Clbit(), True),
    (ClassicalRegister(3, "test_creg"), 3),
    (ClassicalRegister(3, "test_creg"), True),
    expr.lift(Clbit()),
    expr.logic_not(Clbit()),
    expr.equal(ClassicalRegister(3, "test_creg"), 3),
    expr.not_equal(1, ClassicalRegister(3, "test_creg")),
)


@ddt
class TestCreatingControlFlowOperations(QiskitTestCase):
    """Tests instantiation of instruction subclasses for dynamic QuantumCircuits."""

    @idata(CONDITION_PARAMETRISATION)
    def test_while_loop_instantiation(self, condition):
        """Verify creation and properties of a WhileLoopOp."""
        body = QuantumCircuit(3, 1)
        resources = condition_resources(condition)
        body.add_bits(resources.clbits)
        for reg in resources.cregs:
            body.add_register(reg)

        op = WhileLoopOp(condition, body)

        self.assertIsInstance(op, ControlFlowOp)
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "while_loop")
        self.assertEqual(op.num_qubits, body.num_qubits)
        self.assertEqual(op.num_clbits, body.num_clbits)
        self.assertEqual(op.condition, condition)
        self.assertEqual(op.params, [body])
        self.assertEqual(op.blocks, (body,))

    def test_while_loop_invalid_instantiation(self):
        """Verify we catch invalid instantiations of WhileLoopOp."""
        body = QuantumCircuit(3, 1)
        condition = (body.clbits[0], True)

        with self.assertRaisesRegex(CircuitError, r"A classical condition should be a 2-tuple"):
            _ = WhileLoopOp(0, body)

        with self.assertRaisesRegex(CircuitError, r"A classical condition should be a 2-tuple"):
            _ = WhileLoopOp((Clbit(), None), body)

        with self.assertRaisesRegex(CircuitError, r"type 'Bool\(\)'"):
            _ = WhileLoopOp(expr.Value(2, types.Uint(2)), body)

        with self.assertRaisesRegex(CircuitError, r"of type QuantumCircuit"):
            _ = WhileLoopOp(condition, XGate())

    def test_while_loop_invalid_params_setter(self):
        """Verify we catch invalid param settings for WhileLoopOp."""
        body = QuantumCircuit(3, 1)
        condition = (body.clbits[0], True)

        bad_body = QuantumCircuit(2, 1)
        op = WhileLoopOp(condition, body)
        with self.assertRaisesRegex(
            CircuitError, r"num_clbits different than that of the WhileLoopOp"
        ):
            op.params = [bad_body]

    def test_for_loop_iterable_instantiation(self):
        """Verify creation and properties of a ForLoopOp using an iterable indexset."""
        body = QuantumCircuit(3, 1)
        loop_parameter = Parameter("foo")
        indexset = iter(range(0, 10, 2))

        body.rx(loop_parameter, 0)

        op = ForLoopOp(indexset, loop_parameter, body)

        self.assertIsInstance(op, ControlFlowOp)
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "for_loop")
        self.assertEqual(op.num_qubits, 3)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.params, [tuple(range(0, 10, 2)), loop_parameter, body])
        self.assertEqual(op.blocks, (body,))

    def test_for_loop_range_instantiation(self):
        """Verify creation and properties of a ForLoopOp using a range indexset."""
        body = QuantumCircuit(3, 1)
        loop_parameter = Parameter("foo")
        indexset = range(0, 10, 2)

        body.rx(loop_parameter, 0)

        op = ForLoopOp(indexset, loop_parameter, body)

        self.assertIsInstance(op, ControlFlowOp)
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "for_loop")
        self.assertEqual(op.num_qubits, 3)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.params, [indexset, loop_parameter, body])
        self.assertEqual(op.blocks, (body,))

    def test_for_loop_no_parameter_instantiation(self):
        """Verify creation and properties of a ForLoopOp without a loop_parameter."""
        body = QuantumCircuit(3, 1)
        loop_parameter = None
        indexset = range(0, 10, 2)

        body.rx(3.14, 0)

        op = ForLoopOp(indexset, loop_parameter, body)

        self.assertIsInstance(op, ControlFlowOp)
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "for_loop")
        self.assertEqual(op.num_qubits, 3)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.params, [indexset, loop_parameter, body])
        self.assertEqual(op.blocks, (body,))

    def test_for_loop_invalid_instantiation(self):
        """Verify we catch invalid instantiations of ForLoopOp."""
        body = QuantumCircuit(3, 1)
        loop_parameter = Parameter("foo")
        indexset = range(0, 10, 2)

        body.rx(loop_parameter, 0)

        with self.assertWarnsRegex(UserWarning, r"loop_parameter was not found"):
            _ = ForLoopOp(indexset, Parameter("foo"), body)

        with self.assertRaisesRegex(CircuitError, r"to be of type QuantumCircuit"):
            _ = ForLoopOp(indexset, loop_parameter, RXGate(loop_parameter))

    def test_for_loop_invalid_params_setter(self):
        """Verify we catch invalid param settings for ForLoopOp."""
        body = QuantumCircuit(3, 1)
        loop_parameter = Parameter("foo")
        indexset = range(0, 10, 2)

        body.rx(loop_parameter, 0)

        op = ForLoopOp(indexset, loop_parameter, body)

        with self.assertWarnsRegex(UserWarning, r"loop_parameter was not found"):
            op.params = [indexset, Parameter("foo"), body]

        with self.assertRaisesRegex(CircuitError, r"to be of type QuantumCircuit"):
            op.params = [indexset, loop_parameter, RXGate(loop_parameter)]

        bad_body = QuantumCircuit(2, 1)
        with self.assertRaisesRegex(
            CircuitError, r"num_clbits different than that of the ForLoopOp"
        ):
            op.params = [indexset, loop_parameter, bad_body]

        with self.assertRaisesRegex(CircuitError, r"to be either of type Parameter or None"):
            _ = ForLoopOp(indexset, "foo", body)

    @idata(CONDITION_PARAMETRISATION)
    def test_if_else_instantiation_with_else(self, condition):
        """Verify creation and properties of a IfElseOp with an else branch."""
        true_body = QuantumCircuit(3, 1)
        false_body = QuantumCircuit(3, 1)

        op = IfElseOp(condition, true_body, false_body)

        self.assertIsInstance(op, ControlFlowOp)
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "if_else")
        self.assertEqual(op.num_qubits, 3)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.params, [true_body, false_body])
        self.assertEqual(op.condition, condition)
        self.assertEqual(op.blocks, (true_body, false_body))

    @idata(CONDITION_PARAMETRISATION)
    def test_if_else_instantiation_without_else(self, condition):
        """Verify creation and properties of a IfElseOp without an else branch."""
        true_body = QuantumCircuit(3, 1)

        op = IfElseOp(condition, true_body)

        self.assertIsInstance(op, ControlFlowOp)
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "if_else")
        self.assertEqual(op.num_qubits, 3)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.params, [true_body, None])
        self.assertEqual(op.condition, condition)
        self.assertEqual(op.blocks, (true_body,))

    def test_if_else_invalid_instantiation(self):
        """Verify we catch invalid instantiations of IfElseOp."""
        condition = (Clbit(), True)
        true_body = QuantumCircuit(3, 1)
        false_body = QuantumCircuit(3, 1)

        with self.assertRaisesRegex(CircuitError, r"A classical condition should be a 2-tuple"):
            _ = IfElseOp(1, true_body, false_body)

        with self.assertRaisesRegex(CircuitError, r"A classical condition should be a 2-tuple"):
            _ = IfElseOp((1, 2), true_body, false_body)

        with self.assertRaisesRegex(CircuitError, r"type 'Bool\(\)'"):
            _ = IfElseOp(expr.Value(2, types.Uint(2)), true_body, false_body)

        with self.assertRaisesRegex(CircuitError, r"true_body parameter of type QuantumCircuit"):
            _ = IfElseOp(condition, XGate())

        with self.assertRaisesRegex(CircuitError, r"false_body parameter of type QuantumCircuit"):
            _ = IfElseOp(condition, true_body, XGate())

        bad_body = QuantumCircuit(4, 2)

        with self.assertRaisesRegex(
            CircuitError, r"num_clbits different than that of the IfElseOp"
        ):
            _ = IfElseOp(condition, true_body, bad_body)

    def test_if_else_invalid_params_setter(self):
        """Verify we catch invalid param settings for IfElseOp."""
        condition = (Clbit(), True)
        true_body = QuantumCircuit(3, 1)
        false_body = QuantumCircuit(3, 1)

        op = IfElseOp(condition, true_body, false_body)

        with self.assertRaisesRegex(CircuitError, r"true_body parameter of type QuantumCircuit"):
            op.params = [XGate(), None]

        with self.assertRaisesRegex(CircuitError, r"false_body parameter of type QuantumCircuit"):
            op.params = [true_body, XGate()]

        bad_body = QuantumCircuit(4, 2)

        with self.assertRaisesRegex(
            CircuitError, r"num_clbits different than that of the IfElseOp"
        ):
            op.params = [true_body, bad_body]

        with self.assertRaisesRegex(
            CircuitError, r"num_clbits different than that of the IfElseOp"
        ):
            op.params = [bad_body, false_body]

        with self.assertRaisesRegex(
            CircuitError, r"num_clbits different than that of the IfElseOp"
        ):
            op.params = [bad_body, bad_body]

    def test_continue_loop_instantiation(self):
        """Verify creation and properties of a ContinueLoopOp."""
        op = ContinueLoopOp(3, 1)

        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "continue_loop")
        self.assertEqual(op.num_qubits, 3)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.params, [])

    def test_break_loop_instantiation(self):
        """Verify creation and properties of a BreakLoopOp."""
        op = BreakLoopOp(3, 1)

        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "break_loop")
        self.assertEqual(op.num_qubits, 3)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.params, [])

    def test_switch_clbit(self):
        """Test that a switch statement can be constructed with a bit as a condition."""
        qubit = Qubit()
        clbit = Clbit()
        case1 = QuantumCircuit([qubit, clbit])
        case1.x(0)
        case2 = QuantumCircuit([qubit, clbit])
        case2.z(0)

        op = SwitchCaseOp(clbit, [(True, case1), (False, case2)])
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "switch_case")
        self.assertEqual(op.num_qubits, 1)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(op.target, clbit)
        self.assertEqual(op.cases(), {True: case1, False: case2})
        self.assertEqual(list(op.blocks), [case1, case2])

    def test_switch_register(self):
        """Test that a switch statement can be constructed with a register as a condition."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)
        case3 = QuantumCircuit([qubit], creg)
        case3.z(0)

        op = SwitchCaseOp(creg, [(0, case1), (1, case2), (2, case3)])
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "switch_case")
        self.assertEqual(op.num_qubits, 1)
        self.assertEqual(op.num_clbits, 2)
        self.assertEqual(op.target, creg)
        self.assertEqual(op.cases(), {0: case1, 1: case2, 2: case3})
        self.assertEqual(list(op.blocks), [case1, case2, case3])

    def test_switch_expr_uint(self):
        """Test that a switch statement can be constructed with a Uint `Expr` as a condition."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)
        case3 = QuantumCircuit([qubit], creg)
        case3.z(0)

        op = SwitchCaseOp(expr.lift(creg), [(0, case1), (1, case2), (2, case3)])
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "switch_case")
        self.assertEqual(op.num_qubits, 1)
        self.assertEqual(op.num_clbits, 2)
        self.assertEqual(op.target, expr.Var(creg, types.Uint(creg.size)))
        self.assertEqual(op.cases(), {0: case1, 1: case2, 2: case3})
        self.assertEqual(list(op.blocks), [case1, case2, case3])

    def test_switch_expr_bool(self):
        """Test that a switch statement can be constructed with a Bool `Expr` as a condition."""
        qubit = Qubit()
        clbit = Clbit()
        case1 = QuantumCircuit([qubit, clbit])
        case1.x(0)
        case2 = QuantumCircuit([qubit, clbit])
        case2.z(0)

        op = SwitchCaseOp(expr.logic_not(clbit), [(True, case1), (False, case2)])
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "switch_case")
        self.assertEqual(op.num_qubits, 1)
        self.assertEqual(op.num_clbits, 1)
        self.assertEqual(
            op.target,
            expr.Unary(expr.Unary.Op.LOGIC_NOT, expr.Var(clbit, types.Bool()), types.Bool()),
        )
        self.assertEqual(op.cases(), {True: case1, False: case2})
        self.assertEqual(list(op.blocks), [case1, case2])

    def test_switch_with_default(self):
        """Test that a switch statement can be constructed with a default case at the end."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)
        case3 = QuantumCircuit([qubit], creg)
        case3.z(0)

        op = SwitchCaseOp(creg, [(0, case1), (1, case2), (CASE_DEFAULT, case3)])
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "switch_case")
        self.assertEqual(op.num_qubits, 1)
        self.assertEqual(op.num_clbits, 2)
        self.assertEqual(op.target, creg)
        self.assertEqual(op.cases(), {0: case1, 1: case2, CASE_DEFAULT: case3})
        self.assertEqual(list(op.blocks), [case1, case2, case3])

    def test_switch_expr_with_default(self):
        """Test that a switch statement can be constructed with a default case at the end when the
        target is an `Expr`."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)
        case3 = QuantumCircuit([qubit], creg)
        case3.z(0)

        target = expr.bit_xor(creg, 0b11)
        op = SwitchCaseOp(target, [(0, case1), (1, case2), (CASE_DEFAULT, case3)])
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "switch_case")
        self.assertEqual(op.num_qubits, 1)
        self.assertEqual(op.num_clbits, 2)
        self.assertEqual(op.target, target)
        self.assertEqual(op.cases(), {0: case1, 1: case2, CASE_DEFAULT: case3})
        self.assertEqual(list(op.blocks), [case1, case2, case3])

    def test_switch_multiple_cases_to_same_block(self):
        """Test that it is possible to add multiple cases that apply to the same block, if they are
        given as a compound value.  This is an allowed special case of block fall-through."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)

        op = SwitchCaseOp(creg, [(0, case1), ((1, 2), case2)])
        self.assertIsInstance(op, Instruction)
        self.assertEqual(op.name, "switch_case")
        self.assertEqual(op.num_qubits, 1)
        self.assertEqual(op.num_clbits, 2)
        self.assertEqual(op.target, creg)
        self.assertEqual(op.cases(), {0: case1, 1: case2, 2: case2})
        self.assertEqual(list(op.blocks), [case1, case2])

    def test_switch_reconstruction(self):
        """Test that the `cases_specifier` method can be used to reconstruct an equivalent op."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)

        base = SwitchCaseOp(creg, [(0, case1), ((1, 2), case2)])
        self.assertEqual(base, SwitchCaseOp(creg, base.cases_specifier()))

    def test_switch_rejects_separate_cases_to_same_block(self):
        """Test that the switch statement rejects cases that are supplied separately, but point to
        the same QuantumCircuit."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)

        with self.assertRaisesRegex(CircuitError, "ungrouped cases cannot point to the same block"):
            SwitchCaseOp(creg, [(0, case1), (1, case2), (2, case1)])

    def test_switch_rejects_cases_over_different_bits(self):
        """Test that a switch statement fails to build if its individual cases are not all defined
        over the same numbers of bits."""
        qubits = [Qubit() for _ in [None] * 3]
        clbits = [Clbit(), Clbit()]
        case1 = QuantumCircuit(qubits, clbits)
        case2 = QuantumCircuit(qubits[1:], clbits)

        for case in (case1, case2):
            case.h(1)
            case.cx(1, 0)
            case.measure(0, 0)

        with self.assertRaisesRegex(CircuitError, r"incompatible bits between cases"):
            SwitchCaseOp(Clbit(), [(True, case1), (False, case2)])

    def test_switch_rejects_cases_with_bad_types(self):
        """Test that a switch statement will fail to build if it contains cases whose types are not
        matched to the switch expression."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)

        with self.assertRaisesRegex(CircuitError, "case values must be"):
            SwitchCaseOp(creg, [(1.3, case1), (4.5, case2)])

    def test_switch_rejects_cases_after_default(self):
        """Test that a switch statement will fail to build if there are cases after the default
        case."""
        qubit = Qubit()
        creg = ClassicalRegister(2)
        case1 = QuantumCircuit([qubit], creg)
        case1.x(0)
        case2 = QuantumCircuit([qubit], creg)
        case2.y(0)

        with self.assertRaisesRegex(CircuitError, "cases after the default are unreachable"):
            SwitchCaseOp(creg, [(CASE_DEFAULT, case1), (1, case2)])

    def test_if_else_rejects_input_vars(self):
        """Bodies must not contain input variables."""
        cond = (Clbit(), False)
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Bool())
        bad_body = QuantumCircuit(inputs=[a])
        good_body = QuantumCircuit(captures=[a], declarations=[(b, expr.lift(False))])

        with self.assertRaisesRegex(CircuitError, "cannot contain input variables"):
            IfElseOp(cond, bad_body, None)
        with self.assertRaisesRegex(CircuitError, "cannot contain input variables"):
            IfElseOp(cond, bad_body, good_body)
        with self.assertRaisesRegex(CircuitError, "cannot contain input variables"):
            IfElseOp(cond, good_body, bad_body)

    def test_while_rejects_input_vars(self):
        """Bodies must not contain input variables."""
        cond = (Clbit(), False)
        a = expr.Var.new("a", types.Bool())
        bad_body = QuantumCircuit(inputs=[a])
        with self.assertRaisesRegex(CircuitError, "cannot contain input variables"):
            WhileLoopOp(cond, bad_body)

    def test_for_rejects_input_vars(self):
        """Bodies must not contain input variables."""
        a = expr.Var.new("a", types.Bool())
        bad_body = QuantumCircuit(inputs=[a])
        with self.assertRaisesRegex(CircuitError, "cannot contain input variables"):
            ForLoopOp(range(3), None, bad_body)

    def test_switch_rejects_input_vars(self):
        """Bodies must not contain input variables."""
        target = ClassicalRegister(3, "cr")
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Bool())
        bad_body = QuantumCircuit(inputs=[a])
        good_body = QuantumCircuit(captures=[a], declarations=[(b, expr.lift(False))])

        with self.assertRaisesRegex(CircuitError, "cannot contain input variables"):
            SwitchCaseOp(target, [(0, bad_body)])
        with self.assertRaisesRegex(CircuitError, "cannot contain input variables"):
            SwitchCaseOp(target, [(0, good_body), (1, bad_body)])


@ddt
class TestAddingControlFlowOperations(QiskitTestCase):
    """Tests of instruction subclasses for dynamic QuantumCircuits."""

    @data(
        (Clbit(), [False, True]),
        (ClassicalRegister(3, "test_creg"), [3, 1]),
        (ClassicalRegister(3, "test_creg"), [0, (1, 2), CASE_DEFAULT]),
        (expr.lift(Clbit()), [False, True]),
        (expr.lift(ClassicalRegister(3, "test_creg")), [3, 1]),
        (expr.bit_not(ClassicalRegister(3, "test_creg")), [0, (1, 2), CASE_DEFAULT]),
    )
    @unpack
    def test_appending_switch_case_op(self, target, labels):
        """Verify we can append a SwitchCaseOp to a QuantumCircuit."""
        bodies = [QuantumCircuit(3, 1) for _ in labels]

        op = SwitchCaseOp(target, zip(labels, bodies))

        qc = QuantumCircuit(5, 2)
        if isinstance(target, ClassicalRegister):
            qc.add_register(target)
        elif isinstance(target, Clbit):
            qc.add_bits([target])
        else:
            resources = node_resources(target)
            qc.add_bits(resources.clbits)
            for reg in resources.cregs:
                qc.add_register(reg)
        qc.append(op, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "switch_case")
        self.assertEqual(qc.data[0].operation.params, bodies[: len(labels)])
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.data[0].operation.condition, None)
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    @data(
        (Clbit(), [False, True]),
        (ClassicalRegister(3, "test_creg"), [3, 1]),
        (ClassicalRegister(3, "test_creg"), [0, (1, 2), CASE_DEFAULT]),
        (expr.lift(Clbit()), [False, True]),
        (expr.lift(ClassicalRegister(3, "test_creg")), [3, 1]),
        (expr.bit_not(ClassicalRegister(3, "test_creg")), [0, (1, 2), CASE_DEFAULT]),
    )
    @unpack
    def test_quantumcircuit_switch(self, target, labels):
        """Verify we can use the `QuantumCircuit.switch` method."""
        bodies = [QuantumCircuit(3, 1) for _ in labels]

        qc = QuantumCircuit(5, 2)
        if isinstance(target, ClassicalRegister):
            qc.add_register(target)
        elif isinstance(target, Clbit):
            qc.add_bits([target])
        else:
            resources = node_resources(target)
            qc.add_bits(resources.clbits)
            for reg in resources.cregs:
                qc.add_register(reg)
        qc.switch(target, zip(labels, bodies), [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "switch_case")
        self.assertEqual(qc.data[0].operation.params, bodies[: len(labels)])
        self.assertEqual(qc.data[0].operation._condition, None)
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    @idata(CONDITION_PARAMETRISATION)
    def test_appending_while_loop_op(self, condition):
        """Verify we can append a WhileLoopOp to a QuantumCircuit."""
        body = QuantumCircuit(3, 1)

        op = WhileLoopOp(condition, body)

        qc = QuantumCircuit(5, 2)
        qc.append(op, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "while_loop")
        self.assertEqual(qc.data[0].operation.params, [body])
        self.assertEqual(qc.data[0].operation.condition, condition)
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    @idata(CONDITION_PARAMETRISATION)
    def test_quantumcircuit_while_loop(self, condition):
        """Verify we can append a WhileLoopOp to a QuantumCircuit via qc.while_loop."""
        body = QuantumCircuit(3, 1)

        qc = QuantumCircuit(5, 2)
        resources = condition_resources(condition)
        qc.add_bits(resources.clbits)
        for reg in resources.cregs:
            qc.add_register(reg)
        qc.while_loop(condition, body, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "while_loop")
        self.assertEqual(qc.data[0].operation.params, [body])
        self.assertEqual(qc.data[0].operation.condition, condition)
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    def test_appending_for_loop_op(self):
        """Verify we can append a ForLoopOp to a QuantumCircuit."""
        body = QuantumCircuit(3, 1)
        loop_parameter = Parameter("foo")
        indexset = range(0, 10, 2)

        body.rx(loop_parameter, [0, 1, 2])

        op = ForLoopOp(indexset, loop_parameter, body)

        qc = QuantumCircuit(5, 2)
        qc.append(op, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "for_loop")
        self.assertEqual(qc.data[0].operation.params, [indexset, loop_parameter, body])
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    def test_quantumcircuit_for_loop_op(self):
        """Verify we can append a ForLoopOp to a QuantumCircuit via qc.for_loop."""
        body = QuantumCircuit(3, 1)
        loop_parameter = Parameter("foo")
        indexset = range(0, 10, 2)

        body.rx(loop_parameter, [0, 1, 2])

        qc = QuantumCircuit(5, 2)
        qc.for_loop(indexset, loop_parameter, body, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "for_loop")
        self.assertEqual(qc.data[0].operation.params, [indexset, loop_parameter, body])
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    @idata(CONDITION_PARAMETRISATION)
    def test_appending_if_else_op(self, condition):
        """Verify we can append a IfElseOp to a QuantumCircuit."""
        true_body = QuantumCircuit(3, 1)
        false_body = QuantumCircuit(3, 1)

        op = IfElseOp(condition, true_body, false_body)

        qc = QuantumCircuit(5, 2)
        resources = condition_resources(condition)
        qc.add_bits(resources.clbits)
        for reg in resources.cregs:
            qc.add_register(reg)
        qc.append(op, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "if_else")
        self.assertEqual(qc.data[0].operation.params, [true_body, false_body])
        self.assertEqual(qc.data[0].operation.condition, condition)
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    @idata(CONDITION_PARAMETRISATION)
    def test_quantumcircuit_if_else_op(self, condition):
        """Verify we can append a IfElseOp to a QuantumCircuit via qc.if_else."""
        true_body = QuantumCircuit(3, 1)
        false_body = QuantumCircuit(3, 1)

        qc = QuantumCircuit(5, 2)
        resources = condition_resources(condition)
        qc.add_bits(resources.clbits)
        for reg in resources.cregs:
            qc.add_register(reg)
        qc.if_else(condition, true_body, false_body, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "if_else")
        self.assertEqual(qc.data[0].operation.params, [true_body, false_body])
        self.assertEqual(qc.data[0].operation.condition, condition)
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    @idata(CONDITION_PARAMETRISATION)
    def test_quantumcircuit_if_test_op(self, condition):
        """Verify we can append a IfElseOp to a QuantumCircuit via qc.if_test."""
        true_body = QuantumCircuit(3, 1)

        qc = QuantumCircuit(5, 2)
        resources = condition_resources(condition)
        qc.add_bits(resources.clbits)
        for reg in resources.cregs:
            qc.add_register(reg)
        qc.if_test(condition, true_body, [1, 2, 3], [1])

        self.assertEqual(qc.data[0].operation.name, "if_else")
        self.assertEqual(qc.data[0].operation.params, [true_body, None])
        self.assertEqual(qc.data[0].operation.condition, condition)
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits[1:4]))
        self.assertEqual(qc.data[0].clbits, (qc.clbits[1],))

    @idata(CONDITION_PARAMETRISATION)
    def test_appending_if_else_op_with_condition_outside(self, condition):
        """Verify we catch if IfElseOp has a condition outside outer circuit."""
        true_body = QuantumCircuit(3, 1)
        false_body = QuantumCircuit(3, 1)

        qc = QuantumCircuit(5, 2)

        with self.assertRaisesRegex(CircuitError, r".* is not present in this circuit\."):
            qc.if_test(condition, true_body, [1, 2, 3], [1])

        with self.assertRaisesRegex(CircuitError, r".* is not present in this circuit\."):
            qc.if_else(condition, true_body, false_body, [1, 2, 3], [1])

    def test_appending_continue_loop_op(self):
        """Verify we can append a ContinueLoopOp to a QuantumCircuit."""
        op = ContinueLoopOp(3, 1)

        qc = QuantumCircuit(3, 1)
        qc.append(op, [0, 1, 2], [0])

        self.assertEqual(qc.data[0].operation.name, "continue_loop")
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits))
        self.assertEqual(qc.data[0].clbits, tuple(qc.clbits))

    def test_quantumcircuit_continue_loop_op(self):
        """Verify we can append a ContinueLoopOp to a QuantumCircuit via qc.continue_loop."""
        qc = QuantumCircuit(3, 1)
        qc.continue_loop()

        self.assertEqual(qc.data[0].operation.name, "continue_loop")
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits))
        self.assertEqual(qc.data[0].clbits, tuple(qc.clbits))

    def test_appending_break_loop_op(self):
        """Verify we can append a BreakLoopOp to a QuantumCircuit."""
        op = BreakLoopOp(3, 1)

        qc = QuantumCircuit(3, 1)
        qc.append(op, [0, 1, 2], [0])

        self.assertEqual(qc.data[0].operation.name, "break_loop")
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits))
        self.assertEqual(qc.data[0].clbits, tuple(qc.clbits))

    def test_quantumcircuit_break_loop_op(self):
        """Verify we can append a BreakLoopOp to a QuantumCircuit via qc.break_loop."""
        qc = QuantumCircuit(3, 1)
        qc.break_loop()

        self.assertEqual(qc.data[0].operation.name, "break_loop")
        self.assertEqual(qc.data[0].qubits, tuple(qc.qubits))
        self.assertEqual(qc.data[0].clbits, tuple(qc.clbits))

    def test_no_c_if_for_while_loop_if_else(self):
        """Verify we raise if a user attempts to use c_if on an op which sets .condition."""
        qc = QuantumCircuit(3, 1)
        body = QuantumCircuit(1)

        with self.assertRaisesRegex(NotImplementedError, r"cannot be classically controlled"):
            with self.assertWarns(DeprecationWarning):
                qc.while_loop((qc.clbits[0], False), body, [qc.qubits[0]], []).c_if(
                    qc.clbits[0], True
                )

        with self.assertRaisesRegex(NotImplementedError, r"cannot be classically controlled"):
            with self.assertWarns(DeprecationWarning):
                qc.if_test((qc.clbits[0], False), body, [qc.qubits[0]], []).c_if(qc.clbits[0], True)

        with self.assertRaisesRegex(NotImplementedError, r"cannot be classically controlled"):
            with self.assertWarns(DeprecationWarning):
                qc.if_else((qc.clbits[0], False), body, body, [qc.qubits[0]], []).c_if(
                    qc.clbits[0], True
                )

    def test_nested_parameters_are_recognised(self):
        """Verify that parameters added inside a control-flow operator get added to the outer
        circuit table."""
        x, y = Parameter("x"), Parameter("y")

        with self.subTest("if/else"):
            body1 = QuantumCircuit(1, 1)
            body1.rx(x, 0)
            body2 = QuantumCircuit(1, 1)
            body2.rx(y, 0)

            main = QuantumCircuit(1, 1)
            main.if_else((main.clbits[0], 0), body1, body2, [0], [0])
            self.assertEqual({x, y}, set(main.parameters))

        with self.subTest("while"):
            body = QuantumCircuit(1, 1)
            body.rx(x, 0)

            main = QuantumCircuit(1, 1)
            main.while_loop((main.clbits[0], 0), body, [0], [0])
            self.assertEqual({x}, set(main.parameters))

        with self.subTest("for"):
            body = QuantumCircuit(1, 1)
            body.rx(x, 0)

            main = QuantumCircuit(1, 1)
            main.for_loop(range(1), None, body, [0], [0])
            self.assertEqual({x}, set(main.parameters))

    def test_nested_parameters_can_be_assigned(self):
        """Verify that parameters added inside a control-flow operator can be assigned by calls to
        the outer circuit."""
        x, y = Parameter("x"), Parameter("y")

        with self.subTest("if/else"):
            body1 = QuantumCircuit(1, 1)
            body1.rx(x, 0)
            body2 = QuantumCircuit(1, 1)
            body2.rx(y, 0)

            test = QuantumCircuit(1, 1)
            test.if_else((test.clbits[0], 0), body1, body2, [0], [0])
            self.assertEqual({x, y}, set(test.parameters))
            assigned = test.assign_parameters({x: math.pi, y: 0.5 * math.pi})
            self.assertEqual(set(), set(assigned.parameters))

            expected = QuantumCircuit(1, 1)
            expected.if_else(
                (expected.clbits[0], 0),
                body1.assign_parameters({x: math.pi}),
                body2.assign_parameters({y: 0.5 * math.pi}),
                [0],
                [0],
            )

            self.assertEqual(assigned, expected)

        with self.subTest("while"):
            body = QuantumCircuit(1, 1)
            body.rx(x, 0)

            test = QuantumCircuit(1, 1)
            test.while_loop((test.clbits[0], 0), body, [0], [0])
            self.assertEqual({x}, set(test.parameters))
            assigned = test.assign_parameters({x: math.pi})
            self.assertEqual(set(), set(assigned.parameters))

            expected = QuantumCircuit(1, 1)
            expected.while_loop(
                (expected.clbits[0], 0),
                body.assign_parameters({x: math.pi}),
                [0],
                [0],
            )

            self.assertEqual(assigned, expected)

        with self.subTest("for"):
            body = QuantumCircuit(1, 1)
            body.rx(x, 0)

            test = QuantumCircuit(1, 1)
            test.for_loop(range(1), None, body, [0], [0])
            self.assertEqual({x}, set(test.parameters))
            assigned = test.assign_parameters({x: math.pi})
            self.assertEqual(set(), set(assigned.parameters))

            expected = QuantumCircuit(1, 1)
            expected.for_loop(
                range(1),
                None,
                body.assign_parameters({x: math.pi}),
                [0],
                [0],
            )

            self.assertEqual(assigned, expected)

    def test_can_add_op_with_captures_of_inputs(self):
        """Test circuit methods can capture input variables."""
        outer = QuantumCircuit(1, 1)
        a = outer.add_input("a", types.Bool())

        inner = QuantumCircuit(1, 1, captures=[a])

        outer.if_test((outer.clbits[0], False), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "if_else")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.if_else((outer.clbits[0], False), inner.copy(), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "if_else")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})
        self.assertEqual(set(added.blocks[1].iter_captured_vars()), {a})

        outer.while_loop((outer.clbits[0], False), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "while_loop")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.for_loop(range(3), None, inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "for_loop")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.switch(outer.clbits[0], [(False, inner.copy()), (True, inner.copy())], [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "switch_case")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})
        self.assertEqual(set(added.blocks[1].iter_captured_vars()), {a})

    def test_can_add_op_with_captures_of_captures(self):
        """Test circuit methods can capture captured variables."""
        outer = QuantumCircuit(1, 1)
        a = expr.Var.new("a", types.Bool())
        outer.add_capture(a)

        inner = QuantumCircuit(1, 1, captures=[a])

        outer.if_test((outer.clbits[0], False), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "if_else")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.if_else((outer.clbits[0], False), inner.copy(), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "if_else")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})
        self.assertEqual(set(added.blocks[1].iter_captured_vars()), {a})

        outer.while_loop((outer.clbits[0], False), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "while_loop")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.for_loop(range(3), None, inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "for_loop")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.switch(outer.clbits[0], [(False, inner.copy()), (True, inner.copy())], [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "switch_case")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})
        self.assertEqual(set(added.blocks[1].iter_captured_vars()), {a})

    def test_can_add_op_with_captures_of_locals(self):
        """Test circuit methods can capture declared variables."""
        outer = QuantumCircuit(1, 1)
        a = outer.add_var("a", expr.lift(True))

        inner = QuantumCircuit(1, 1, captures=[a])

        outer.if_test((outer.clbits[0], False), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "if_else")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.if_else((outer.clbits[0], False), inner.copy(), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "if_else")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})
        self.assertEqual(set(added.blocks[1].iter_captured_vars()), {a})

        outer.while_loop((outer.clbits[0], False), inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "while_loop")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.for_loop(range(3), None, inner.copy(), [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "for_loop")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})

        outer.switch(outer.clbits[0], [(False, inner.copy()), (True, inner.copy())], [0], [0])
        added = outer.data[-1].operation
        self.assertEqual(added.name, "switch_case")
        self.assertEqual(set(added.blocks[0].iter_captured_vars()), {a})
        self.assertEqual(set(added.blocks[1].iter_captured_vars()), {a})

    def test_cannot_capture_unknown_variables_methods(self):
        """Control-flow operations should not be able to capture variables that don't exist in the
        outer circuit."""
        outer = QuantumCircuit(1, 1)

        a = expr.Var.new("a", types.Bool())
        inner = QuantumCircuit(1, 1, captures=[a])

        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.if_test((outer.clbits[0], False), inner.copy(), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.if_else((outer.clbits[0], False), inner.copy(), inner.copy(), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.while_loop((outer.clbits[0], False), inner.copy(), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.for_loop(range(3), None, inner.copy(), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.switch(outer.clbits[0], [(False, inner.copy()), (True, inner.copy())], [0], [0])

    def test_cannot_capture_unknown_variables_append(self):
        """Control-flow operations should not be able to capture variables that don't exist in the
        outer circuit."""
        outer = QuantumCircuit(1, 1)

        a = expr.Var.new("a", types.Bool())
        inner = QuantumCircuit(1, 1, captures=[a])

        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.append(IfElseOp((outer.clbits[0], False), inner.copy(), None), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.append(IfElseOp((outer.clbits[0], False), inner.copy(), inner.copy()), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.append(WhileLoopOp((outer.clbits[0], False), inner.copy()), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.append(ForLoopOp(range(3), None, inner.copy()), [0], [0])
        with self.assertRaisesRegex(CircuitError, "not in this circuit"):
            outer.append(
                SwitchCaseOp(outer.clbits[0], [(False, inner.copy()), (True, inner.copy())]),
                [0],
                [0],
            )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Translates gates to a target basis using a given equivalence library."""

import logging

from collections import defaultdict

from qiskit.transpiler.basepasses import TransformationPass
from qiskit._accelerate.basis.basis_translator import base_run

logger = logging.getLogger(__name__)


class BasisTranslator(TransformationPass):
    """Translates gates to a target basis by searching for a set of translations
    from a given EquivalenceLibrary.

    This pass operates in several steps:

    * Determine the source basis from the input circuit.
    * Perform a Dijkstra search over basis sets, starting from the device's
      target_basis new gates are being generated using the rules from the provided
      EquivalenceLibrary and the search stops if all gates in the source basis have
      been generated.
    * The found path, as a set of rules from the EquivalenceLibrary, is composed
      into a set of gate replacement rules.
    * The composed replacement rules are applied in-place to each op node which
      is not already in the target_basis.

    If the target keyword argument is specified and that
    :class:`~qiskit.transpiler.Target` objects contains operations
    which are non-global (i.e. they are defined only for a subset of qubits),
    as calculated by :meth:`~qiskit.transpiler.Target.get_non_global_operation_names`,
    this pass will attempt to match the output translation to those constraints.
    For 1 qubit operations this is straightforward, the pass will perform a
    search using the union of the set of global operations with the set of operations
    defined solely on that qubit. For multi-qubit gates this is a bit more involved,
    while the behavior is initially similar to the single qubit case, just using all
    the qubits the operation is run on (where order is not significant) isn't sufficient.
    We also need to consider any potential local qubits defined on subsets of the
    quantum arguments for the multi-qubit operation. This means the target used for the
    search of a non-global multi-qubit gate is the union of global operations, non-global
    multi-qubit gates sharing the same qubits, and any non-global gates defined on
    any subset of the qubits used.


    .. note::

        In the case of non-global operations it is possible for a single
        execution of this pass to output an incomplete translation if any
        non-global gates are defined on qubits that are a subset of a larger
        multi-qubit gate. For example, if you have a ``u`` gate only defined on
        qubit 0 and an ``x`` gate only on qubit 1 it is possible when
        translating a 2 qubit operation on qubit 0 and 1 that the output might
        have ``u`` on qubit 1 and ``x`` on qubit 0. Typically running this pass
        a second time will correct these issues.

    .. _translation_errors:

    Translation Errors
    ------------------

    This pass will error if there is no path to translate an input gate to
    the specified basis. However, during a typical/default preset passmanager
    this pass gets run multiple times at different stages of the compilation
    pipeline. This means that potentially the input gates that are getting
    translated were not in the input circuit to :func:`~.transpile` as they
    were generated by an intermediate transform in the circuit.

    When this error occurs it typically means that either the target basis
    is not universal or there are additional equivalence rules needed in the
    :class:`~.EquivalenceLibrary` instance being used by the
    :class:`~.BasisTranslator` pass. You can refer to
    :ref:`custom_basis_gates` for details on adding custom equivalence rules.
    """

    def __init__(self, equivalence_library, target_basis, target=None, min_qubits=0):
        """Initialize a BasisTranslator instance.

        Args:
            equivalence_library (EquivalenceLibrary): The equivalence library
                which will be used by the BasisTranslator pass. (Instructions in
                this library will not be unrolled by this pass.)
            target_basis (list[str]): Target basis names to unroll to, e.g. ``['u3', 'cx']``.
            target (Target): The backend compilation target
            min_qubits (int): The minimum number of qubits for operations in the input
                dag to translate.
        """

        super().__init__()
        self._equiv_lib = equivalence_library
        self._target_basis = target_basis
        self._target = target
        self._non_global_operations = None
        self._qargs_with_non_global_operation = {}
        self._min_qubits = min_qubits
        if target is not None:
            self._non_global_operations = self._target.get_non_global_operation_names()
            self._qargs_with_non_global_operation = defaultdict(set)
            for gate in self._non_global_operations:
                for qarg in self._target[gate]:
                    self._qargs_with_non_global_operation[qarg].add(gate)

    def run(self, dag):
        """Translate an input DAGCircuit to the target basis.

        Args:
            dag (DAGCircuit): input dag

        Raises:
            TranspilerError: if the target basis cannot be reached

        Returns:
            DAGCircuit: translated circuit.
        """

        return base_run(
            dag,
            self._equiv_lib,
            self._qargs_with_non_global_operation,
            self._min_qubits,
            None if self._target_basis is None else set(self._target_basis),
            self._target,
            None if self._non_global_operations is None else set(self._non_global_operations),
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=redefined-builtin

"""Object to represent the information at a node in the DAGCircuit."""
from __future__ import annotations

from qiskit.exceptions import QiskitError


class DAGDepNode:
    """Object to represent the information at a node in the DAGDependency().

    It is used as the return value from `*_nodes()` functions and can
    be supplied to functions that take a node.
    """

    __slots__ = [
        "type",
        "_op",
        "name",
        "_qargs",
        "cargs",
        "sort_key",
        "node_id",
        "successors",
        "predecessors",
        "reachable",
        "matchedwith",
        "isblocked",
        "successorstovisit",
        "qindices",
        "cindices",
    ]

    def __init__(
        self,
        type=None,
        op=None,
        name=None,
        qargs=(),
        cargs=(),
        successors: list[int] | None = None,
        predecessors: list[int] | None = None,
        reachable=None,
        matchedwith: list[int] | None = None,
        successorstovisit: list[int] | None = None,
        isblocked: bool | None = None,
        qindices: list[int] | None = None,
        cindices: list[int] | None = None,
        nid: int = -1,
    ):

        self.type = type
        self._op = op
        self.name = name
        self._qargs = tuple(qargs) if qargs is not None else ()
        self.cargs = tuple(cargs) if cargs is not None else ()
        self.node_id = nid
        self.sort_key: str = str(self._qargs)
        self.successors: list[int] = successors if successors is not None else []
        self.predecessors: list[int] = predecessors if predecessors is not None else []
        self.reachable = reachable
        self.matchedwith: list[int] = matchedwith if matchedwith is not None else []
        self.isblocked: bool = isblocked
        self.successorstovisit: list[int] = (
            successorstovisit if successorstovisit is not None else []
        )
        self.qindices: list[int] = qindices if qindices is not None else []
        self.cindices: list[int] = cindices if cindices is not None else []

    @property
    def op(self):
        """Returns the Instruction object corresponding to the op for the node, else None"""
        if not self.type or self.type != "op":
            raise QiskitError(f"The node {str(self)} is not an op node")
        return self._op

    @op.setter
    def op(self, data):
        self._op = data

    @property
    def qargs(self):
        """
        Returns list of Qubit, else an empty list.
        """
        return self._qargs

    @qargs.setter
    def qargs(self, new_qargs):
        """Sets the qargs to be the given list of qargs."""
        self._qargs = tuple(new_qargs)
        self.sort_key = str(new_qargs)

    @staticmethod
    def semantic_eq(node1, node2):
        """
        Check if DAG nodes are considered equivalent, e.g., as a node_match for nx.is_isomorphic.

        Args:
            node1 (DAGDepNode): A node to compare.
            node2 (DAGDepNode): The other node to compare.

        Return:
            Bool: If node1 == node2
        """
        # For barriers, qarg order is not significant so compare as sets
        if "barrier" == node1.name == node2.name:
            return set(node1._qargs) == set(node2._qargs)

        if node1.type == node2.type:
            if node1._op == node2._op:
                if node1.name == node2.name:
                    if node1._qargs == node2._qargs:
                        if node1.cargs == node2.cargs:
                            if node1.type == "op":
                                if getattr(node1._op, "condition", None) != getattr(
                                    node2._op, "condition", None
                                ):
                                    return False
                            return True
        return False

    def copy(self):
        """
        Function to copy a DAGDepNode object.
        Returns:
            DAGDepNode: a copy of a DAGDepNode object.
        """

        dagdepnode = DAGDepNode()

        dagdepnode.type = self.type
        dagdepnode._op = self.op
        dagdepnode.name = self.name
        dagdepnode._qargs = self._qargs
        dagdepnode.cargs = self.cargs
        dagdepnode.node_id = self.node_id
        dagdepnode.sort_key = self.sort_key
        dagdepnode.successors = self.successors
        dagdepnode.predecessors = self.predecessors
        dagdepnode.reachable = self.reachable
        dagdepnode.isblocked = self.isblocked
        dagdepnode.successorstovisit = self.successorstovisit
        dagdepnode.qindices = self.qindices
        dagdepnode.cindices = self.cindices
        dagdepnode.matchedwith = self.matchedwith.copy()

        return dagdepnode



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the LayoutTransformation pass"""

import unittest

from qiskit import QuantumRegister, QuantumCircuit
from qiskit.converters import circuit_to_dag
from qiskit.transpiler import CouplingMap, Layout, Target
from qiskit.circuit.library import CXGate
from qiskit.transpiler.passes import LayoutTransformation
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestLayoutTransformation(QiskitTestCase):
    """
    Tests the LayoutTransformation pass.
    """

    def test_three_qubit(self):
        """Test if the permutation {0->2,1->0,2->1} is implemented correctly."""
        v = QuantumRegister(3, "v")  # virtual qubits
        coupling = CouplingMap([[0, 1], [1, 2]])
        from_layout = Layout({v[0]: 0, v[1]: 1, v[2]: 2})
        to_layout = Layout({v[0]: 2, v[1]: 0, v[2]: 1})
        ltpass = LayoutTransformation(
            coupling_map=coupling, from_layout=from_layout, to_layout=to_layout, seed=42
        )
        qc = QuantumCircuit(3)
        dag = circuit_to_dag(qc)
        output_dag = ltpass.run(dag)

        expected = QuantumCircuit(3)
        expected.swap(1, 0)
        expected.swap(1, 2)

        self.assertEqual(circuit_to_dag(expected), output_dag)

    def test_four_qubit(self):
        """Test if the permutation {0->3,1->0,2->1,3->2} is implemented correctly."""
        v = QuantumRegister(4, "v")  # virtual qubits
        coupling = CouplingMap([[0, 1], [1, 2], [2, 3]])
        from_layout = Layout({v[0]: 0, v[1]: 1, v[2]: 2, v[3]: 3})
        to_layout = Layout({v[0]: 3, v[1]: 0, v[2]: 1, v[3]: 2})
        ltpass = LayoutTransformation(
            coupling_map=coupling, from_layout=from_layout, to_layout=to_layout, seed=42
        )
        qc = QuantumCircuit(4)  # input (empty) physical circuit
        dag = circuit_to_dag(qc)
        output_dag = ltpass.run(dag)

        expected = QuantumCircuit(4)
        expected.swap(1, 0)
        expected.swap(1, 2)
        expected.swap(2, 3)

        self.assertEqual(circuit_to_dag(expected), output_dag)

    def test_four_qubit_with_target(self):
        """Test if the permutation {0->3,1->0,2->1,3->2} is implemented correctly."""
        v = QuantumRegister(4, "v")  # virtual qubits
        target = Target()
        target.add_instruction(CXGate(), {(0, 1): None, (1, 2): None, (2, 3): None})
        from_layout = Layout({v[0]: 0, v[1]: 1, v[2]: 2, v[3]: 3})
        to_layout = Layout({v[0]: 3, v[1]: 0, v[2]: 1, v[3]: 2})
        ltpass = LayoutTransformation(target, from_layout=from_layout, to_layout=to_layout, seed=42)
        qc = QuantumCircuit(4)  # input (empty) physical circuit
        dag = circuit_to_dag(qc)
        output_dag = ltpass.run(dag)

        expected = QuantumCircuit(4)
        expected.swap(1, 0)
        expected.swap(1, 2)
        expected.swap(2, 3)

        self.assertEqual(circuit_to_dag(expected), output_dag)

    def test_full_connected_coupling_map(self):
        """Test if the permutation {0->3,1->0,2->1,3->2} in a fully connected map."""
        v = QuantumRegister(4, "v")  # virtual qubits
        from_layout = Layout({v[0]: 0, v[1]: 1, v[2]: 2, v[3]: 3})
        to_layout = Layout({v[0]: 3, v[1]: 0, v[2]: 1, v[3]: 2})
        ltpass = LayoutTransformation(
            coupling_map=None, from_layout=from_layout, to_layout=to_layout, seed=42
        )
        qc = QuantumCircuit(4)  # input (empty) physical circuit
        dag = circuit_to_dag(qc)
        output_dag = ltpass.run(dag)

        expected = QuantumCircuit(4)
        expected.swap(1, 0)
        expected.swap(2, 1)
        expected.swap(3, 2)

        self.assertEqual(circuit_to_dag(expected), output_dag)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-docstring,invalid-name,no-member
# pylint: disable=attribute-defined-outside-init

import os

import qiskit


class TranspilerBenchSuite:
    def _build_cx_circuit(self):
        cx_register = qiskit.QuantumRegister(2)
        cx_circuit = qiskit.QuantumCircuit(cx_register)
        cx_circuit.h(cx_register[0])
        cx_circuit.h(cx_register[0])
        cx_circuit.cx(cx_register[0], cx_register[1])
        cx_circuit.cx(cx_register[0], cx_register[1])
        cx_circuit.cx(cx_register[0], cx_register[1])
        cx_circuit.cx(cx_register[0], cx_register[1])
        return cx_circuit

    def _build_single_gate_circuit(self):
        single_register = qiskit.QuantumRegister(1)
        single_gate_circuit = qiskit.QuantumCircuit(single_register)
        single_gate_circuit.h(single_register[0])
        return single_gate_circuit

    def setup(self):
        self.single_gate_circuit = self._build_single_gate_circuit()
        self.cx_circuit = self._build_cx_circuit()
        self.qasm_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "qasm"))
        large_qasm_path = os.path.join(self.qasm_path, "test_eoh_qasm.qasm")
        self.large_qasm = qiskit.QuantumCircuit.from_qasm_file(large_qasm_path)
        self.coupling_map = [
            [0, 1],
            [1, 0],
            [1, 2],
            [1, 4],
            [2, 1],
            [2, 3],
            [3, 2],
            [3, 5],
            [4, 1],
            [4, 7],
            [5, 3],
            [5, 8],
            [6, 7],
            [7, 4],
            [7, 6],
            [7, 10],
            [8, 5],
            [8, 9],
            [8, 11],
            [9, 8],
            [10, 7],
            [10, 12],
            [11, 8],
            [11, 14],
            [12, 10],
            [12, 13],
            [12, 15],
            [13, 12],
            [13, 14],
            [14, 11],
            [14, 13],
            [14, 16],
            [15, 12],
            [15, 18],
            [16, 14],
            [16, 19],
            [17, 18],
            [18, 15],
            [18, 17],
            [18, 21],
            [19, 16],
            [19, 20],
            [19, 22],
            [20, 19],
            [21, 18],
            [21, 23],
            [22, 19],
            [22, 25],
            [23, 21],
            [23, 24],
            [24, 23],
            [24, 25],
            [25, 22],
            [25, 24],
            [25, 26],
            [26, 25],
        ]
        self.basis = ["id", "rz", "sx", "x", "cx", "reset"]

    def time_single_gate_compile(self):
        circ = qiskit.compiler.transpile(
            self.single_gate_circuit,
            coupling_map=self.coupling_map,
            basis_gates=self.basis,
            seed_transpiler=20220125,
        )
        qiskit.compiler.assemble(circ)

    def time_cx_compile(self):
        circ = qiskit.compiler.transpile(
            self.cx_circuit,
            coupling_map=self.coupling_map,
            basis_gates=self.basis,
            seed_transpiler=20220125,
        )
        qiskit.compiler.assemble(circ)

    def time_compile_from_large_qasm(self):
        circ = qiskit.compiler.transpile(
            self.large_qasm,
            coupling_map=self.coupling_map,
            basis_gates=self.basis,
            seed_transpiler=20220125,
        )
        qiskit.compiler.assemble(circ)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""Tests for Weyl coordinate routines."""

import unittest
import numpy as np
from numpy.testing import assert_allclose

from qiskit._accelerate.two_qubit_decompose import weyl_coordinates
from qiskit.quantum_info.random import random_unitary
from qiskit.synthesis.two_qubit.local_invariance import (
    two_qubit_local_invariants,
    local_equivalence,
)
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestWeyl(QiskitTestCase):
    """Test Weyl coordinate routines"""

    def test_weyl_coordinates_simple(self):
        """Check Weyl coordinates against known cases."""
        # Identity [0,0,0]
        U = np.identity(4, dtype=complex)
        weyl = weyl_coordinates(U)
        assert_allclose(weyl, [0, 0, 0])

        # CNOT [pi/4, 0, 0]
        U = np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]], dtype=complex)
        weyl = weyl_coordinates(U)
        assert_allclose(weyl, [np.pi / 4, 0, 0], atol=1e-07)

        # SWAP [pi/4, pi/4 ,pi/4]
        U = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=complex)

        weyl = weyl_coordinates(U)
        assert_allclose(weyl, [np.pi / 4, np.pi / 4, np.pi / 4])

        # SQRT ISWAP [pi/8, pi/8, 0]
        U = np.array(
            [
                [1, 0, 0, 0],
                [0, 1 / np.sqrt(2), 1j / np.sqrt(2), 0],
                [0, 1j / np.sqrt(2), 1 / np.sqrt(2), 0],
                [0, 0, 0, 1],
            ],
            dtype=complex,
        )

        weyl = weyl_coordinates(U)
        assert_allclose(weyl, [np.pi / 8, np.pi / 8, 0])

    def test_weyl_coordinates_random(self):
        """Randomly check Weyl coordinates with local invariants."""
        for _ in range(10):
            U = random_unitary(4).data
            weyl = weyl_coordinates(U)
            local_equiv = local_equivalence(weyl)
            local = two_qubit_local_invariants(U)
            assert_allclose(local, local_equiv)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=invalid-name,missing-docstring
# pylint: disable=attribute-defined-outside-init

from qiskit import transpile
from qiskit.circuit.library.standard_gates import XGate
from qiskit.transpiler import CouplingMap, PassManager
from qiskit.transpiler import InstructionDurations
from qiskit.transpiler.passes import (
    TimeUnitConversion,
    ASAPScheduleAnalysis,
    ALAPScheduleAnalysis,
    PadDynamicalDecoupling,
)
from qiskit.converters import circuit_to_dag

from .utils import random_circuit


class SchedulingPassBenchmarks:

    params = ([5, 10, 20], [500, 1000])
    param_names = ["n_qubits", "depth"]
    timeout = 300

    def setup(self, n_qubits, depth):
        seed = 42
        self.circuit = random_circuit(
            n_qubits, depth, measure=True, conditional=False, reset=False, seed=seed, max_operands=2
        )
        self.basis_gates = ["rz", "sx", "x", "cx", "id", "reset"]
        self.cmap = [
            [0, 1],
            [1, 0],
            [1, 2],
            [1, 6],
            [2, 1],
            [2, 3],
            [3, 2],
            [3, 4],
            [3, 8],
            [4, 3],
            [5, 6],
            [5, 10],
            [6, 1],
            [6, 5],
            [6, 7],
            [7, 6],
            [7, 8],
            [7, 12],
            [8, 3],
            [8, 7],
            [8, 9],
            [9, 8],
            [9, 14],
            [10, 5],
            [10, 11],
            [11, 10],
            [11, 12],
            [11, 16],
            [12, 7],
            [12, 11],
            [12, 13],
            [13, 12],
            [13, 14],
            [13, 18],
            [14, 9],
            [14, 13],
            [15, 16],
            [16, 11],
            [16, 15],
            [16, 17],
            [17, 16],
            [17, 18],
            [18, 13],
            [18, 17],
            [18, 19],
            [19, 18],
        ]
        self.coupling_map = CouplingMap(self.cmap)
        self.transpiled_circuit = transpile(
            self.circuit,
            basis_gates=self.basis_gates,
            coupling_map=self.coupling_map,
            optimization_level=1,
        )
        self.dag = circuit_to_dag(self.transpiled_circuit)
        self.durations = InstructionDurations(
            [
                ("rz", None, 0),
                ("id", None, 160),
                ("sx", None, 160),
                ("x", None, 160),
                ("cx", None, 800),
                ("measure", None, 3200),
                ("reset", None, 3600),
            ],
            dt=1e-9,
        )

    def time_time_unit_conversion_pass(self, _, __):
        TimeUnitConversion(self.durations).run(self.dag)

    def time_alap_schedule_pass(self, _, __):
        dd_sequence = [XGate(), XGate()]
        pm = PassManager(
            [
                ALAPScheduleAnalysis(self.durations),
                PadDynamicalDecoupling(self.durations, dd_sequence),
            ]
        )
        pm.run(self.transpiled_circuit)

    def time_asap_schedule_pass(self, _, __):
        dd_sequence = [XGate(), XGate()]
        pm = PassManager(
            [
                ASAPScheduleAnalysis(self.durations),
                PadDynamicalDecoupling(self.durations, dd_sequence),
            ]
        )
        pm.run(self.transpiled_circuit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
===========================================
Circuit Scheduler (:mod:`qiskit.scheduler`)
===========================================

.. currentmodule:: qiskit.scheduler

A circuit scheduler compiles a circuit program to a pulse program.

Core API
========

.. autoclass:: ScheduleConfig

.. currentmodule:: qiskit.scheduler.schedule_circuit
.. autofunction:: schedule_circuit
.. currentmodule:: qiskit.scheduler

Pulse scheduling methods
========================

.. currentmodule:: qiskit.scheduler.methods
.. autofunction:: as_soon_as_possible
.. autofunction:: as_late_as_possible
.. currentmodule:: qiskit.scheduler
"""
from qiskit.scheduler import schedule_circuit
from qiskit.scheduler.config import ScheduleConfig



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
A collection of functions that decide the layout of an output image.
See :py:mod:`~qiskit.visualization.timeline.types` for more info on the required data.

There are 2 types of layout functions in this module.

1. layout.bit_arrange

In this stylesheet entry the input data is a list of `types.Bits` and returns a
sorted list of `types.Bits`.


The function signature of the layout is restricted to:

    ```python

    def my_layout(bits: List[types.Bits]) -> List[types.Bits]:

        # your code here: sort input bits and return list of bits
    ```

2. layout.time_axis_map

In this stylesheet entry the input data is `Tuple[int, int]` that represents horizontal
axis limit of the output image. The layout function returns `types.HorizontalAxis` data
which is consumed by the plotter API to make horizontal axis.

The function signature of the layout is restricted to:

    ```python

    def my_layout(time_window: Tuple[int, int]) -> types.HorizontalAxis:

        # your code here: create and return axis config
    ```

Arbitrary layout function satisfying the above format can be accepted.
"""
from typing import List, Tuple
import numpy as np

from qiskit import circuit
from qiskit.visualization.timeline import types


def qreg_creg_ascending(bits: List[types.Bits]) -> List[types.Bits]:
    """Sort bits by ascending order.

    Bit order becomes Q0, Q1, ..., Cl0, Cl1, ...

    Args:
        bits: List of bits to sort.

    Returns:
        Sorted bits.
    """
    return [x for x in bits if isinstance(x, circuit.Qubit)] + [
        x for x in bits if isinstance(x, circuit.Clbit)
    ]


def qreg_creg_descending(bits: List[types.Bits]) -> List[types.Bits]:
    """Sort bits by descending order.

    Bit order becomes Q_N, Q_N-1, ..., Cl_N, Cl_N-1, ...

    Args:
        bits: List of bits to sort.

    Returns:
        Sorted bits.
    """
    return [x for x in bits[::-1] if isinstance(x, circuit.Qubit)] + [
        x for x in bits[::-1] if isinstance(x, circuit.Clbit)
    ]


def time_map_in_dt(time_window: Tuple[int, int]) -> types.HorizontalAxis:
    """Layout function for the horizontal axis formatting.

    Generate equispaced 6 horizontal axis ticks.

    Args:
        time_window: Left and right edge of this graph.

    Returns:
        Axis formatter object.
    """
    # shift time axis
    t0, t1 = time_window

    # axis label
    axis_loc = np.linspace(max(t0, 0), t1, 6)
    axis_label = axis_loc.copy()

    # consider time resolution
    label = "System cycle time (dt)"

    formatted_label = [f"{val:.0f}" for val in axis_label]

    return types.HorizontalAxis(
        window=(t0, t1), axis_map=dict(zip(axis_loc, formatted_label)), label=label
    )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-class-docstring,missing-function-docstring
# pylint: disable=missing-module-docstring

import datetime
import itertools
import operator
import unittest
import warnings

from test import combine
from ddt import ddt, data

from qiskit.circuit import QuantumCircuit
from qiskit.compiler import assemble
from qiskit.compiler import transpile
from qiskit.exceptions import QiskitError
from qiskit.providers.fake_provider import (
    Fake5QV1,
    Fake20QV1,
    Fake7QPulseV1,
    Fake27QPulseV1,
    Fake127QPulseV1,
    FakeOpenPulse2Q,
    GenericBackendV2,
)
from qiskit.providers.backend_compat import BackendV2Converter, convert_to_target
from qiskit.providers.models.backendproperties import BackendProperties
from qiskit.providers.models.backendconfiguration import GateConfig
from qiskit.providers.backend import BackendV2
from qiskit.utils import optionals
from qiskit.circuit.library import (
    SXGate,
    MCPhaseGate,
    MCXGate,
    RZGate,
    RXGate,
    U2Gate,
    U1Gate,
    U3Gate,
    YGate,
    ZGate,
    PauliGate,
    SwapGate,
    RGate,
    MCXGrayCode,
    RYGate,
    CZGate,
    ECRGate,
    UnitaryGate,
    UCGate,
    Initialize,
    DiagonalGate,
)
from qiskit.circuit import ControlledGate, Parameter
from qiskit.quantum_info.operators.channel.quantum_channel import QuantumChannel
from qiskit.quantum_info.operators.channel.kraus import Kraus
from qiskit.quantum_info.operators.channel import SuperOp
from qiskit.circuit.controlflow import (
    IfElseOp,
    WhileLoopOp,
    ForLoopOp,
    ContinueLoopOp,
    BreakLoopOp,
    SwitchCaseOp,
)
from qiskit.transpiler.coupling import CouplingMap
from test.utils.base import QiskitTestCase  # pylint: disable=wrong-import-order

with warnings.catch_warnings():
    BACKENDS = [Fake5QV1(), Fake20QV1(), Fake7QPulseV1(), Fake27QPulseV1(), Fake127QPulseV1()]

BACKENDS_V2 = []
for n in [5, 7, 16, 20, 27, 65, 127]:
    cmap = CouplingMap.from_ring(n)
    BACKENDS_V2.append(GenericBackendV2(num_qubits=n, coupling_map=cmap, seed=42))


@ddt
class TestFakeBackends(QiskitTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.circuit = QuantumCircuit(2)
        cls.circuit.h(0)
        cls.circuit.h(1)
        cls.circuit.h(0)
        cls.circuit.h(1)
        cls.circuit.x(0)
        cls.circuit.x(1)
        cls.circuit.measure_all()

    @combine(
        backend=BACKENDS_V2,
        optimization_level=[0, 1, 2, 3],
    )
    def test_circuit_on_fake_backend_v2(self, backend, optimization_level):
        if not optionals.HAS_AER and backend.num_qubits > 20:
            self.skipTest(f"Unable to run fake_backend {backend.name} without qiskit-aer")
        job = backend.run(
            transpile(
                self.circuit, backend, seed_transpiler=42, optimization_level=optimization_level
            ),
            seed_simulator=42,
        )
        result = job.result()
        counts = result.get_counts()
        max_count = max(counts.items(), key=operator.itemgetter(1))[0]
        self.assertEqual(max_count, "11")

    @combine(
        backend=BACKENDS,
        optimization_level=[0, 1, 2, 3],
        dsc="Test execution path on {backend} with optimization level {optimization_level}",
        name="{backend}_opt_level_{optimization_level}",
    )
    def test_circuit_on_fake_backend(self, backend, optimization_level):
        if not optionals.HAS_AER and backend.configuration().num_qubits > 20:
            self.skipTest(
                f"Unable to run fake_backend {backend.configuration().backend_name} without qiskit-aer"
            )
        with self.assertWarnsRegex(
            DeprecationWarning,
            expected_regex="The `transpile` function will "
            "stop supporting inputs of type `BackendV1`",
        ):
            transpiled = transpile(
                self.circuit, backend, seed_transpiler=42, optimization_level=optimization_level
            )
        job = backend.run(transpiled, seed_simulator=42)
        result = job.result()
        counts = result.get_counts()
        max_count = max(counts.items(), key=operator.itemgetter(1))[0]
        self.assertEqual(max_count, "11")

    def test_qobj_failure(self):
        backend = BACKENDS[-1]
        with self.assertWarns(DeprecationWarning):
            tqc = transpile(self.circuit, backend)
            qobj = assemble(tqc, backend)
        with self.assertRaises(QiskitError):
            backend.run(qobj)

    @data(*BACKENDS)
    def test_to_dict_properties(self, backend):
        with warnings.catch_warnings():
            # The class QobjExperimentHeader is deprecated
            warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
            properties = backend.properties()
        if properties:
            self.assertIsInstance(backend.properties().to_dict(), dict)
        else:
            self.assertTrue(backend.configuration().simulator)

    @data(*BACKENDS_V2)
    def test_convert_to_target(self, backend):
        target = backend.target
        if target.dt is not None:
            self.assertLess(target.dt, 1e-6)

    @data(*BACKENDS_V2)
    def test_backend_v2_dtm(self, backend):
        if backend.dtm:
            self.assertLess(backend.dtm, 1e-6)

    @data(*BACKENDS)
    def test_to_dict_configuration(self, backend):
        configuration = backend.configuration()
        if configuration.open_pulse:
            self.assertLess(configuration.dt, 1e-6)
            self.assertLess(configuration.dtm, 1e-6)
            for i in configuration.qubit_lo_range:
                self.assertGreater(i[0], 1e6)
                self.assertGreater(i[1], 1e6)
                self.assertLess(i[0], i[1])

            for i in configuration.meas_lo_range:
                self.assertGreater(i[0], 1e6)
                self.assertGreater(i[0], 1e6)
                self.assertLess(i[0], i[1])

            for i in configuration.rep_times:
                self.assertGreater(i, 0)
                self.assertLess(i, 1)

        self.assertIsInstance(configuration.to_dict(), dict)

    @data(*BACKENDS)
    def test_defaults_to_dict(self, backend):
        if hasattr(backend, "defaults"):
            defaults = backend.defaults()
            self.assertIsInstance(defaults.to_dict(), dict)

            for i in defaults.qubit_freq_est:
                self.assertGreater(i, 1e6)
                self.assertGreater(i, 1e6)

            for i in defaults.meas_freq_est:
                self.assertGreater(i, 1e6)
                self.assertGreater(i, 1e6)
        else:
            self.skipTest(f"Backend {backend} does not have defaults")

    def test_delay_circuit(self):
        with self.assertWarns(DeprecationWarning):
            backend = Fake27QPulseV1()
        backend.configuration().timing_constraints = {
            "acquire_alignment": 1,
            "granularity": 1,
            "min_length": 1,
            "pulse_alignment": 1,
        }
        qc = QuantumCircuit(2)
        qc.delay(502, 0, unit="ns")
        qc.x(1)
        qc.delay(250, 1, unit="ns")
        qc.measure_all()
        with self.assertWarnsRegex(
            DeprecationWarning,
            expected_regex="The `transpile` function will "
            "stop supporting inputs of type `BackendV1`",
        ):
            res = transpile(qc, backend)
        self.assertIn("delay", res.count_ops())

    @data(0, 1, 2, 3)
    def test_converter(self, opt_level):
        with self.assertWarns(DeprecationWarning):
            backend_v1 = Fake5QV1()
            backend_v2 = BackendV2Converter(backend_v1)
        self.assertIsInstance(backend_v2, BackendV2)
        res = transpile(self.circuit, backend_v2, optimization_level=opt_level)
        job = backend_v2.run(res)
        with warnings.catch_warnings():
            # TODO remove this catch once Aer stops using QobjDictField
            warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
            result = job.result()
        counts = result.get_counts()
        max_count = max(counts.items(), key=operator.itemgetter(1))[0]
        self.assertEqual(max_count, "11")

    def test_converter_delay_circuit(self):
        with self.assertWarns(DeprecationWarning):
            backend = Fake27QPulseV1()
        backend.configuration().timing_constraints = {
            "acquire_alignment": 1,
            "granularity": 1,
            "min_length": 1,
            "pulse_alignment": 1,
        }
        with self.assertWarns(DeprecationWarning):
            backend_v2 = BackendV2Converter(backend, add_delay=True)
        self.assertIsInstance(backend_v2, BackendV2)
        qc = QuantumCircuit(2)
        qc.delay(502, 0, unit="ns")
        qc.x(1)
        qc.delay(250, 1, unit="ns")
        qc.measure_all()
        res = transpile(qc, backend_v2)
        self.assertIn("delay", res.count_ops())

    def test_converter_with_missing_gate_property(self):
        """Test converting to V2 model with irregular backend data."""
        with self.assertWarns(DeprecationWarning):
            backend = FakeOpenPulse2Q()
        # The backend includes pulse calibration definition for U2, but its property is gone.
        # Note that u2 is a basis gate of this device.
        # Since gate property is not provided, the gate broadcasts to all qubits as ideal instruction.
        del backend._properties._gates["u2"]

        with self.assertWarns(DeprecationWarning):
            # This should not raise error
            backend_v2 = BackendV2Converter(backend, add_delay=True)
        self.assertDictEqual(backend_v2.target["u2"], {None: None})

    def test_non_cx_tests(self):
        backend = GenericBackendV2(num_qubits=5, basis_gates=["cz", "x", "sx", "id", "rz"], seed=42)
        self.assertIsInstance(backend.target.operation_from_name("cz"), CZGate)
        backend = GenericBackendV2(
            num_qubits=5, basis_gates=["ecr", "x", "sx", "id", "rz"], seed=42
        )
        self.assertIsInstance(backend.target.operation_from_name("ecr"), ECRGate)

    @unittest.skipUnless(optionals.HAS_AER, "Aer required for this test")
    def test_converter_simulator(self):
        class MCSXGate(ControlledGate):
            def __init__(self, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcsx",
                    1 + num_ctrl_qubits,
                    [],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=SXGate(),
                )

        class MCYGate(ControlledGate):
            def __init__(self, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcy",
                    1 + num_ctrl_qubits,
                    [],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=YGate(),
                )

        class MCZGate(ControlledGate):
            def __init__(self, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcz",
                    1 + num_ctrl_qubits,
                    [],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=ZGate(),
                )

        class MCRXGate(ControlledGate):
            def __init__(self, theta, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcrx",
                    1 + num_ctrl_qubits,
                    [theta],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=RXGate(theta),
                )

        class MCRYGate(ControlledGate):
            def __init__(self, theta, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcry",
                    1 + num_ctrl_qubits,
                    [theta],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=RYGate(theta),
                )

        class MCRZGate(ControlledGate):
            def __init__(self, theta, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcrz",
                    1 + num_ctrl_qubits,
                    [theta],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=RZGate(theta),
                )

        class MCRGate(ControlledGate):
            def __init__(self, theta, phi, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcr",
                    1 + num_ctrl_qubits,
                    [theta, phi],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=RGate(theta, phi),
                )

        class MCU1Gate(ControlledGate):
            def __init__(self, theta, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcu1",
                    1 + num_ctrl_qubits,
                    [theta],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=U1Gate(theta),
                )

        class MCU2Gate(ControlledGate):
            def __init__(self, theta, lam, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcu2",
                    1 + num_ctrl_qubits,
                    [theta, lam],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=U2Gate(theta, lam),
                )

        class MCU3Gate(ControlledGate):
            def __init__(self, theta, lam, phi, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcu3",
                    1 + num_ctrl_qubits,
                    [theta, phi, lam],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=U3Gate(theta, phi, lam),
                )

        class MCUGate(ControlledGate):
            def __init__(self, theta, lam, phi, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcu",
                    1 + num_ctrl_qubits,
                    [theta, phi, lam],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=U3Gate(theta, phi, lam),
                )

        class MCSwapGate(ControlledGate):
            def __init__(self, num_ctrl_qubits, ctrl_state=None):
                super().__init__(
                    "mcswap",
                    2 + num_ctrl_qubits,
                    [],
                    None,
                    num_ctrl_qubits,
                    ctrl_state=ctrl_state,
                    base_gate=SwapGate(),
                )

        from qiskit_aer import AerSimulator
        from qiskit_aer.library import (
            SaveExpectationValue,
            SaveAmplitudes,
            SaveStatevectorDict,
            SaveSuperOp,
            SaveClifford,
            SaveMatrixProductState,
            SaveDensityMatrix,
            SaveProbabilities,
            SaveStatevector,
            SetDensityMatrix,
            SetUnitary,
            SaveState,
            SetMatrixProductState,
            SaveUnitary,
            SetSuperOp,
            SaveExpectationValueVariance,
            SaveStabilizer,
            SetStatevector,
            SetStabilizer,
            SaveAmplitudesSquared,
            SaveProbabilitiesDict,
        )
        from qiskit_aer.noise.errors import ReadoutError
        from qiskit_aer.noise.noise_model import QuantumErrorLocation

        sim = AerSimulator()
        # test only if simulator's backend is V1
        if sim.version > 1:
            return
        phi = Parameter("phi")
        lam = Parameter("lam")
        backend = BackendV2Converter(
            sim,
            name_mapping={
                "mcsx": MCSXGate,
                "mcp": MCPhaseGate,
                "mcphase": MCPhaseGate,
                "quantum_channel": QuantumChannel,
                "initialize": Initialize,
                "save_expval": SaveExpectationValue,
                "diagonal": DiagonalGate,
                "save_amplitudes": SaveAmplitudes,
                "roerror": ReadoutError,
                "mcrx": MCRXGate,
                "kraus": Kraus,
                "save_statevector_dict": SaveStatevectorDict,
                "mcx": MCXGate,
                "mcu1": MCU1Gate,
                "mcu2": MCU2Gate,
                "mcu3": MCU3Gate,
                "save_superop": SaveSuperOp,
                "multiplexer": UCGate,
                "mcy": MCYGate,
                "superop": SuperOp,
                "save_clifford": SaveClifford,
                "save_matrix_product_state": SaveMatrixProductState,
                "save_density_matrix": SaveDensityMatrix,
                "save_probabilities": SaveProbabilities,
                "if_else": IfElseOp,
                "while_loop": WhileLoopOp,
                "for_loop": ForLoopOp,
                "switch_case": SwitchCaseOp,
                "break_loop": BreakLoopOp,
                "continue_loop": ContinueLoopOp,
                "save_statevector": SaveStatevector,
                "mcu": MCUGate,
                "set_density_matrix": SetDensityMatrix,
                "qerror_loc": QuantumErrorLocation,
                "unitary": UnitaryGate,
                "mcz": MCZGate,
                "pauli": PauliGate,
                "set_unitary": SetUnitary,
                "save_state": SaveState,
                "mcswap": MCSwapGate,
                "set_matrix_product_state": SetMatrixProductState,
                "save_unitary": SaveUnitary,
                "mcr": MCRGate,
                "mcx_gray": MCXGrayCode,
                "mcrz": MCRZGate,
                "set_superop": SetSuperOp,
                "save_expval_var": SaveExpectationValueVariance,
                "save_stabilizer": SaveStabilizer,
                "set_statevector": SetStatevector,
                "mcry": MCRYGate,
                "set_stabilizer": SetStabilizer,
                "save_amplitudes_sq": SaveAmplitudesSquared,
                "save_probabilities_dict": SaveProbabilitiesDict,
                "cu2": U2Gate(phi, lam).control(),
            },
        )
        self.assertIsInstance(backend, BackendV2)
        res = transpile(self.circuit, backend)
        job = backend.run(res)
        result = job.result()
        counts = result.get_counts()
        max_count = max(counts.items(), key=operator.itemgetter(1))[0]
        self.assertEqual(max_count, "11")

    def test_filter_faulty_qubits_backend_v2_converter(self):
        """Test faulty qubits in v2 conversion."""
        with self.assertWarns(DeprecationWarning):
            backend = Fake127QPulseV1()
            # Get properties dict to make it easier to work with the properties API
            # is difficult to edit because of the multiple layers of nesting and
            # different object types
            props_dict = backend.properties().to_dict()
        for i in range(62, 67):
            non_operational = {
                "date": datetime.datetime.now(datetime.timezone.utc),
                "name": "operational",
                "unit": "",
                "value": 0,
            }
            props_dict["qubits"][i].append(non_operational)

        with self.assertWarns(DeprecationWarning):
            backend._properties = BackendProperties.from_dict(props_dict)
            v2_backend = BackendV2Converter(backend, filter_faulty=True)
        for i in range(62, 67):
            for qarg in v2_backend.target.qargs:
                self.assertNotIn(i, qarg)

    def test_filter_faulty_qubits_backend_v2_converter_with_delay(self):
        """Test faulty qubits in v2 conversion."""
        with self.assertWarns(DeprecationWarning):
            backend = Fake127QPulseV1()
            # Get properties dict to make it easier to work with the properties API
            # is difficult to edit because of the multiple layers of nesting and
            # different object types
            props_dict = backend.properties().to_dict()
        for i in range(62, 67):
            non_operational = {
                "date": datetime.datetime.now(datetime.timezone.utc),
                "name": "operational",
                "unit": "",
                "value": 0,
            }
            props_dict["qubits"][i].append(non_operational)
        with self.assertWarns(DeprecationWarning):
            backend._properties = BackendProperties.from_dict(props_dict)
            v2_backend = BackendV2Converter(backend, filter_faulty=True, add_delay=True)
        for i in range(62, 67):
            for qarg in v2_backend.target.qargs:
                self.assertNotIn(i, qarg)

    def test_backend_v2_converter_without_delay(self):
        """Test setting :code:`add_delay`argument of :func:`.BackendV2Converter`
        to :code:`False`."""

        expected = {
            (0,),
            (0, 1),
            (0, 2),
            (1,),
            (1, 0),
            (1, 2),
            (2,),
            (2, 0),
            (2, 1),
            (2, 3),
            (2, 4),
            (3,),
            (3, 2),
            (3, 4),
            (4,),
            (4, 2),
            (4, 3),
        }
        with self.assertWarns(DeprecationWarning):
            backend_v1 = Fake5QV1()
            backend_v2 = BackendV2Converter(backend=backend_v1, filter_faulty=True, add_delay=False)

        self.assertEqual(backend_v2.target.qargs, expected)

    def test_backend_v2_converter_with_meaningless_gate_config(self):
        """Test backend with broken gate config can be converted only with properties data."""
        with self.assertWarns(DeprecationWarning):
            backend_v1 = Fake5QV1()

            backend_v1.configuration().gates = [
                GateConfig(name="NotValidGate", parameters=[], qasm_def="not_valid_gate")
            ]
            backend_v2 = BackendV2Converter(
                backend=backend_v1,
                filter_faulty=True,
                add_delay=False,
            )
        ops_with_measure = backend_v2.target.operation_names
        self.assertCountEqual(
            ops_with_measure,
            backend_v1.configuration().basis_gates + ["measure"],
        )

    def test_filter_faulty_qubits_and_gates_backend_v2_converter(self):
        """Test faulty gates and qubits."""
        with self.assertWarns(DeprecationWarning):
            backend = Fake127QPulseV1()
            # Get properties dict to make it easier to work with the properties API
            # is difficult to edit because of the multiple layers of nesting and
            # different object types
            props_dict = backend.properties().to_dict()
        for i in range(62, 67):
            non_operational = {
                "date": datetime.datetime.now(datetime.timezone.utc),
                "name": "operational",
                "unit": "",
                "value": 0,
            }
            props_dict["qubits"][i].append(non_operational)
        invalid_cx_edges = {
            (113, 114),
            (114, 113),
            (96, 100),
            (100, 96),
            (114, 109),
            (109, 114),
            (24, 34),
            (34, 24),
        }
        non_operational_gate = {
            "date": datetime.datetime.now(datetime.timezone.utc),
            "name": "operational",
            "unit": "",
            "value": 0,
        }
        for gate in props_dict["gates"]:
            if tuple(gate["qubits"]) in invalid_cx_edges:
                gate["parameters"].append(non_operational_gate)

        with self.assertWarns(DeprecationWarning):
            backend._properties = BackendProperties.from_dict(props_dict)
            v2_backend = BackendV2Converter(backend, filter_faulty=True)
        for i in range(62, 67):
            for qarg in v2_backend.target.qargs:
                self.assertNotIn(i, qarg)
        for edge in invalid_cx_edges:
            self.assertNotIn(edge, v2_backend.target["cx"])

    def test_filter_faulty_gates_v2_converter(self):
        """Test just faulty gates in conversion."""
        with self.assertWarns(DeprecationWarning):
            backend = Fake127QPulseV1()
        # Get properties dict to make it easier to work with the properties API
        # is difficult to edit because of the multiple layers of nesting and
        # different object types
        props_dict = backend.properties().to_dict()
        invalid_cx_edges = {
            (113, 114),
            (114, 113),
            (96, 100),
            (100, 96),
            (114, 109),
            (109, 114),
            (24, 34),
            (34, 24),
        }
        non_operational_gate = {
            "date": datetime.datetime.now(datetime.timezone.utc),
            "name": "operational",
            "unit": "",
            "value": 0,
        }
        for gate in props_dict["gates"]:
            if tuple(gate["qubits"]) in invalid_cx_edges:
                gate["parameters"].append(non_operational_gate)

        with self.assertWarns(DeprecationWarning):
            backend._properties = BackendProperties.from_dict(props_dict)
            v2_backend = BackendV2Converter(backend, filter_faulty=True)
        for i in range(62, 67):
            self.assertIn((i,), v2_backend.target.qargs)
        for edge in invalid_cx_edges:
            self.assertNotIn(edge, v2_backend.target["cx"])

    def test_filter_faulty_no_faults_v2_converter(self):
        """Test that faulty qubit filtering does nothing with all operational qubits and gates."""
        with self.assertWarns(DeprecationWarning):
            backend = Fake127QPulseV1()
            v2_backend = BackendV2Converter(backend, filter_faulty=True)
        for i in range(v2_backend.num_qubits):
            self.assertIn((i,), v2_backend.target.qargs)

    @data(0, 1, 2, 3)
    def test_faulty_full_path_transpile_connected_cmap(self, opt_level):
        with self.assertWarns(DeprecationWarning):
            backend = Fake5QV1()
            props = backend.properties().to_dict()

        non_operational_gate = {
            "date": datetime.datetime.now(datetime.timezone.utc),
            "name": "operational",
            "unit": "",
            "value": 0,
        }
        for gate in props["gates"]:
            if tuple(sorted(gate["qubits"])) == (0, 1):
                gate["parameters"].append(non_operational_gate)
        with self.assertWarns(DeprecationWarning):
            backend._properties = BackendProperties.from_dict(props)
            v2_backend = BackendV2Converter(backend, filter_faulty=True)
        qc = QuantumCircuit(5)
        for x, y in itertools.product(range(5), range(5)):
            if x == y:
                continue
            qc.cx(x, y)
        tqc = transpile(qc, v2_backend, seed_transpiler=433, optimization_level=opt_level)
        connections = [tuple(sorted(tqc.find_bit(q).index for q in x.qubits)) for x in tqc.data]
        self.assertNotIn((0, 1), connections)

    def test_convert_to_target_control_flow(self):
        with self.assertWarns(DeprecationWarning):
            backend = Fake27QPulseV1()
        properties = backend.properties()
        configuration = backend.configuration()
        configuration.supported_instructions = [
            "cx",
            "id",
            "delay",
            "measure",
            "reset",
            "rz",
            "sx",
            "x",
            "if_else",
            "for_loop",
            "switch_case",
        ]
        defaults = backend.defaults()
        with self.assertWarns(DeprecationWarning):
            target = convert_to_target(configuration, properties, defaults)
        self.assertTrue(target.instruction_supported("if_else", ()))
        self.assertFalse(target.instruction_supported("while_loop", ()))
        self.assertTrue(target.instruction_supported("for_loop", ()))
        self.assertTrue(target.instruction_supported("switch_case", ()))

    def test_convert_unrelated_supported_instructions(self):
        with self.assertWarns(DeprecationWarning):
            backend = Fake27QPulseV1()
        properties = backend.properties()
        configuration = backend.configuration()
        configuration.supported_instructions = [
            "cx",
            "id",
            "delay",
            "measure",
            "reset",
            "rz",
            "sx",
            "x",
            "play",
            "u2",
            "u3",
            "u1",
            "shiftf",
            "acquire",
            "setf",
            "if_else",
            "for_loop",
            "switch_case",
        ]
        defaults = backend.defaults()
        with self.assertWarns(DeprecationWarning):
            target = convert_to_target(configuration, properties, defaults)
        self.assertTrue(target.instruction_supported("if_else", ()))
        self.assertFalse(target.instruction_supported("while_loop", ()))
        self.assertTrue(target.instruction_supported("for_loop", ()))
        self.assertTrue(target.instruction_supported("switch_case", ()))
        self.assertFalse(target.instruction_supported("u3", (0,)))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utils for reading a user preference config files."""

import configparser
import os
from warnings import warn

from qiskit import exceptions

DEFAULT_FILENAME = os.path.join(os.path.expanduser("~"), ".qiskit", "settings.conf")


class UserConfig:
    """Class representing a user config file

    The config file format should look like:

    [default]
    circuit_drawer = mpl
    circuit_mpl_style = default
    circuit_mpl_style_path = ~/.qiskit:<default location>
    circuit_reverse_bits = True
    circuit_idle_wires = False
    transpile_optimization_level = 1
    parallel = False
    num_processes = 4
    sabre_all_threads = true

    """

    def __init__(self, filename=None):
        """Create a UserConfig

        Args:
            filename (str): The path to the user config file. If one isn't
                specified, ~/.qiskit/settings.conf is used.
        """
        if filename is None:
            self.filename = DEFAULT_FILENAME
        else:
            self.filename = filename
        self.settings = {}
        self.config_parser = configparser.ConfigParser()

    def read_config_file(self):
        """Read config file and parse the contents into the settings attr."""
        if not os.path.isfile(self.filename):
            return
        self.config_parser.read(self.filename)
        if "default" in self.config_parser.sections():
            # Parse circuit_drawer
            circuit_drawer = self.config_parser.get("default", "circuit_drawer", fallback=None)
            if circuit_drawer:
                if circuit_drawer not in ["text", "mpl", "latex", "latex_source", "auto"]:
                    raise exceptions.QiskitUserConfigError(
                        f"{circuit_drawer} is not a valid circuit drawer backend. Must be "
                        "either 'text', 'mpl', 'latex', 'latex_source', or "
                        "'auto'."
                    )
                self.settings["circuit_drawer"] = circuit_drawer

            # Parse state_drawer
            state_drawer = self.config_parser.get("default", "state_drawer", fallback=None)
            if state_drawer:
                valid_state_drawers = [
                    "repr",
                    "text",
                    "latex",
                    "latex_source",
                    "qsphere",
                    "hinton",
                    "bloch",
                ]
                if state_drawer not in valid_state_drawers:
                    valid_choices_string = "', '".join(c for c in valid_state_drawers)
                    raise exceptions.QiskitUserConfigError(
                        f"'{state_drawer}' is not a valid state drawer backend. "
                        f"Choose from: '{valid_choices_string}'"
                    )
                self.settings["state_drawer"] = state_drawer

            # Parse circuit_mpl_style
            circuit_mpl_style = self.config_parser.get(
                "default", "circuit_mpl_style", fallback=None
            )
            if circuit_mpl_style:
                if not isinstance(circuit_mpl_style, str):
                    warn(
                        f"{circuit_mpl_style} is not a valid mpl circuit style. Must be "
                        "a text string. Will not load style.",
                        UserWarning,
                        2,
                    )
                self.settings["circuit_mpl_style"] = circuit_mpl_style

            # Parse circuit_mpl_style_path
            circuit_mpl_style_path = self.config_parser.get(
                "default", "circuit_mpl_style_path", fallback=None
            )
            if circuit_mpl_style_path:
                cpath_list = circuit_mpl_style_path.split(":")
                for path in cpath_list:
                    if not os.path.exists(os.path.expanduser(path)):
                        warn(
                            f"{path} is not a valid circuit mpl style path."
                            " Correct the path in ~/.qiskit/settings.conf.",
                            UserWarning,
                            2,
                        )
                self.settings["circuit_mpl_style_path"] = cpath_list

            # Parse circuit_reverse_bits
            try:
                circuit_reverse_bits = self.config_parser.getboolean(
                    "default", "circuit_reverse_bits", fallback=None
                )
            except ValueError as err:
                raise exceptions.QiskitUserConfigError(
                    f"Value assigned to circuit_reverse_bits is not valid. {str(err)}"
                )
            if circuit_reverse_bits is not None:
                self.settings["circuit_reverse_bits"] = circuit_reverse_bits

            # Parse circuit_idle_wires
            try:
                circuit_idle_wires = self.config_parser.getboolean(
                    "default", "circuit_idle_wires", fallback=None
                )
            except ValueError as err:
                raise exceptions.QiskitUserConfigError(
                    f"Value assigned to circuit_idle_wires is not valid. {str(err)}"
                )
            if circuit_idle_wires is not None:
                self.settings["circuit_idle_wires"] = circuit_idle_wires

            # Parse transpile_optimization_level
            transpile_optimization_level = self.config_parser.getint(
                "default", "transpile_optimization_level", fallback=-1
            )
            if transpile_optimization_level != -1:
                if transpile_optimization_level < 0 or transpile_optimization_level > 3:
                    raise exceptions.QiskitUserConfigError(
                        "%s is not a valid optimization level. Must be 0, 1, 2, or 3."
                    )
                self.settings["transpile_optimization_level"] = transpile_optimization_level

            # Parse parallel
            parallel_enabled = self.config_parser.getboolean("default", "parallel", fallback=None)
            if parallel_enabled is not None:
                self.settings["parallel_enabled"] = parallel_enabled

            # Parse num_processes
            num_processes = self.config_parser.getint("default", "num_processes", fallback=-1)
            if num_processes != -1:
                if num_processes <= 0:
                    raise exceptions.QiskitUserConfigError(
                        "%s is not a valid number of processes. Must be greater than 0"
                    )
                self.settings["num_processes"] = num_processes

            # Parse sabre_all_threads
            sabre_all_threads = self.config_parser.getboolean(
                "default", "sabre_all_threads", fallback=None
            )
            if sabre_all_threads is not None:
                self.settings["sabre_all_threads"] = sabre_all_threads


def set_config(key, value, section=None, file_path=None):
    """Adds or modifies a user configuration

    It will add configuration to the currently configured location
    or the value of file argument.

    Only valid user config can be set in 'default' section. Custom
    user config can be added in any other sections.

    Changes to the existing config file will not be reflected in
    the current session since the config file is parsed at import time.

    Args:
        key (str): name of the config
        value (obj): value of the config
        section (str, optional): if not specified, adds it to the
            `default` section of the config file.
        file_path (str, optional): the file to which config is added.
            If not specified, adds it to the default config file or
            if set, the value of `QISKIT_SETTINGS` env variable.

    Raises:
        QiskitUserConfigError: if the config is invalid
    """
    filename = file_path or os.getenv("QISKIT_SETTINGS", DEFAULT_FILENAME)
    section = "default" if section is None else section

    if not isinstance(key, str):
        raise exceptions.QiskitUserConfigError("Key must be string type")

    valid_config = {
        "circuit_drawer",
        "circuit_mpl_style",
        "circuit_mpl_style_path",
        "circuit_reverse_bits",
        "circuit_idle_wires",
        "transpile_optimization_level",
        "parallel",
        "num_processes",
        "sabre_all_threads",
    }

    if section in [None, "default"]:
        if key not in valid_config:
            raise exceptions.QiskitUserConfigError(f"{key} is not a valid user config.")

    config = configparser.ConfigParser()
    config.read(filename)

    if section not in config.sections():
        config.add_section(section)

    config.set(section, key, str(value))

    try:
        with open(filename, "w") as cfgfile:
            config.write(cfgfile)
    except OSError as ex:
        raise exceptions.QiskitUserConfigError(
            f"Unable to load the config file {filename}. Error: '{str(ex)}'"
        )

    # validates config
    user_config = UserConfig(filename)
    user_config.read_config_file()


def get_config():
    """Read the config file from the default location or env var

    It will read a config file at either the default location
    ~/.qiskit/settings.conf or if set the value of the QISKIT_SETTINGS env var.

    It will return the parsed settings dict from the parsed config file.
    Returns:
        dict: The settings dict from the parsed config file.
    """
    filename = os.getenv("QISKIT_SETTINGS", DEFAULT_FILENAME)
    if not os.path.isfile(filename):
        return {}
    user_config = UserConfig(filename)
    user_config.read_config_file()
    return user_config.settings



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Test Qiskit's QuantumCircuit class for multiple registers."""
from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit
from qiskit.converters.circuit_to_dag import circuit_to_dag
from qiskit.circuit.exceptions import CircuitError
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestCircuitMultiRegs(QiskitTestCase):
    """QuantumCircuit Qasm tests."""

    def test_circuit_multi(self):
        """Test circuit multi regs declared at start."""
        qreg0 = QuantumRegister(2, "q0")
        creg0 = ClassicalRegister(2, "c0")
        qreg1 = QuantumRegister(2, "q1")
        creg1 = ClassicalRegister(2, "c1")
        circ = QuantumCircuit(qreg0, qreg1, creg0, creg1)
        circ.x(qreg0[1])
        circ.x(qreg1[0])

        meas = QuantumCircuit(qreg0, qreg1, creg0, creg1)
        meas.measure(qreg0, creg0)
        meas.measure(qreg1, creg1)

        qc = circ.compose(meas)

        circ2 = QuantumCircuit()
        circ2.add_register(qreg0)
        circ2.add_register(qreg1)
        circ2.add_register(creg0)
        circ2.add_register(creg1)
        circ2.x(qreg0[1])
        circ2.x(qreg1[0])

        meas2 = QuantumCircuit()
        meas2.add_register(qreg0)
        meas2.add_register(qreg1)
        meas2.add_register(creg0)
        meas2.add_register(creg1)
        meas2.measure(qreg0, creg0)
        meas2.measure(qreg1, creg1)

        qc2 = circ2.compose(meas2)

        dag_qc = circuit_to_dag(qc)
        dag_qc2 = circuit_to_dag(qc2)
        dag_circ2 = circuit_to_dag(circ2)
        dag_circ = circuit_to_dag(circ)

        self.assertEqual(dag_qc, dag_qc2)
        self.assertEqual(dag_circ, dag_circ2)

    def test_circuit_multi_name_collision(self):
        """Test circuit multi regs, with name collision."""
        qreg0 = QuantumRegister(2, "q")
        qreg1 = QuantumRegister(3, "q")
        self.assertRaises(CircuitError, QuantumCircuit, qreg0, qreg1)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test basic simulator."""

import os
import unittest
import numpy as np

from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister
from qiskit.compiler import transpile
from qiskit.providers.basic_provider import BasicSimulator
from qiskit.qasm2 import dumps
from test import QiskitTestCase  # pylint: disable=wrong-import-order


from . import BasicProviderBackendTestMixin


class TestBasicSimulator(QiskitTestCase, BasicProviderBackendTestMixin):
    """Test the basic provider simulator."""

    def setUp(self):
        super().setUp()
        self.backend = BasicSimulator()
        bell = QuantumCircuit(2, 2)
        bell.h(0)
        bell.cx(0, 1)
        bell.measure([0, 1], [0, 1])
        self.circuit = bell

        self.seed = 88
        self.backend = BasicSimulator()
        qasm_dir = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))), "qasm"
        )
        qasm_filename = os.path.join(qasm_dir, "example.qasm")
        qcirc = QuantumCircuit.from_qasm_file(qasm_filename)
        qcirc.name = "test"
        self.transpiled_circuit = transpile(qcirc, backend=self.backend)

    def test_basic_simulator_single_shot(self):
        """Test single shot run."""
        shots = 1
        result = self.backend.run(
            self.transpiled_circuit, shots=shots, seed_simulator=self.seed
        ).result()
        self.assertEqual(result.success, True)

    def test_measure_sampler_repeated_qubits(self):
        """Test measure sampler if qubits measured more than once."""
        shots = 100
        qr = QuantumRegister(2, "qr")
        cr = ClassicalRegister(4, "cr")
        circuit = QuantumCircuit(qr, cr)
        circuit.x(qr[1])
        circuit.measure(qr[0], cr[0])
        circuit.measure(qr[1], cr[1])
        circuit.measure(qr[1], cr[2])
        circuit.measure(qr[0], cr[3])
        target = {"0110": shots}
        job = self.backend.run(
            transpile(circuit, self.backend), shots=shots, seed_simulator=self.seed
        )
        result = job.result()
        counts = result.get_counts(0)
        self.assertEqual(counts, target)

    def test_measure_sampler_single_qubit(self):
        """Test measure sampler if single-qubit is measured."""
        shots = 100
        num_qubits = 5
        qr = QuantumRegister(num_qubits, "qr")
        cr = ClassicalRegister(1, "cr")

        for qubit in range(num_qubits):
            circuit = QuantumCircuit(qr, cr)
            circuit.x(qr[qubit])
            circuit.measure(qr[qubit], cr[0])
            target = {"1": shots}
            job = self.backend.run(
                transpile(circuit, self.backend), shots=shots, seed_simulator=self.seed
            )
            result = job.result()
            counts = result.get_counts(0)
            self.assertEqual(counts, target)

    def test_measure_sampler_partial_qubit(self):
        """Test measure sampler if single-qubit is measured."""
        shots = 100
        num_qubits = 5
        qr = QuantumRegister(num_qubits, "qr")
        cr = ClassicalRegister(4, "cr")

        #             ░     ░     ░ ┌─┐ ░
        # qr_0: ──────░─────░─────░─┤M├─░────
        #       ┌───┐ ░     ░ ┌─┐ ░ └╥┘ ░
        # qr_1: ┤ X ├─░─────░─┤M├─░──╫──░────
        #       └───┘ ░     ░ └╥┘ ░  ║  ░
        # qr_2: ──────░─────░──╫──░──╫──░────
        #       ┌───┐ ░ ┌─┐ ░  ║  ░  ║  ░ ┌─┐
        # qr_3: ┤ X ├─░─┤M├─░──╫──░──╫──░─┤M├
        #       └───┘ ░ └╥┘ ░  ║  ░  ║  ░ └╥┘
        # qr_4: ──────░──╫──░──╫──░──╫──░──╫─
        #             ░  ║  ░  ║  ░  ║  ░  ║
        # cr: 4/═════════╩═════╩═════╩═════╩═
        #                1     0     2     3
        circuit = QuantumCircuit(qr, cr)
        circuit.x(qr[3])
        circuit.x(qr[1])
        circuit.barrier(qr)
        circuit.measure(qr[3], cr[1])
        circuit.barrier(qr)
        circuit.measure(qr[1], cr[0])
        circuit.barrier(qr)
        circuit.measure(qr[0], cr[2])
        circuit.barrier(qr)
        circuit.measure(qr[3], cr[3])
        target = {"1011": shots}
        job = self.backend.run(
            transpile(circuit, self.backend), shots=shots, seed_simulator=self.seed
        )
        result = job.result()
        counts = result.get_counts(0)
        self.assertEqual(counts, target)

    def test_basic_simulator(self):
        """Test data counts output for single circuit run against reference."""
        result = self.backend.run(
            self.transpiled_circuit, shots=1000, seed_simulator=self.seed
        ).result()
        shots = 1024
        threshold = 0.04 * shots
        counts = result.get_counts("test")
        target = {
            "100 100": shots / 8,
            "011 011": shots / 8,
            "101 101": shots / 8,
            "111 111": shots / 8,
            "000 000": shots / 8,
            "010 010": shots / 8,
            "110 110": shots / 8,
            "001 001": shots / 8,
        }
        self.assertDictAlmostEqual(counts, target, threshold)

    def test_if_statement(self):
        """Test if statements."""
        shots = 100
        qr = QuantumRegister(3, "qr")
        cr = ClassicalRegister(3, "cr")

        #       ┌───┐┌─┐          ┌─┐
        # qr_0: ┤ X ├┤M├──────────┤M├──────
        #       ├───┤└╥┘┌─┐       └╥┘┌─┐
        # qr_1: ┤ X ├─╫─┤M├────────╫─┤M├───
        #       └───┘ ║ └╥┘ ┌───┐  ║ └╥┘┌─┐
        # qr_2: ──────╫──╫──┤ X ├──╫──╫─┤M├
        #             ║  ║  └─╥─┘  ║  ║ └╥┘
        #             ║  ║ ┌──╨──┐ ║  ║  ║
        # cr: 3/══════╩══╩═╡ 0x3 ╞═╩══╩══╩═
        #             0  1 └─────┘ 0  1  2
        circuit_if_true = QuantumCircuit(qr, cr)
        circuit_if_true.x(qr[0])
        circuit_if_true.x(qr[1])
        circuit_if_true.measure(qr[0], cr[0])
        circuit_if_true.measure(qr[1], cr[1])
        with self.assertWarns(DeprecationWarning):
            circuit_if_true.x(qr[2]).c_if(cr, 0x3)
        circuit_if_true.measure(qr[0], cr[0])
        circuit_if_true.measure(qr[1], cr[1])
        circuit_if_true.measure(qr[2], cr[2])

        #       ┌───┐┌─┐       ┌─┐
        # qr_0: ┤ X ├┤M├───────┤M├──────
        #       └┬─┬┘└╥┘       └╥┘┌─┐
        # qr_1: ─┤M├──╫─────────╫─┤M├───
        #        └╥┘  ║  ┌───┐  ║ └╥┘┌─┐
        # qr_2: ──╫───╫──┤ X ├──╫──╫─┤M├
        #         ║   ║  └─╥─┘  ║  ║ └╥┘
        #         ║   ║ ┌──╨──┐ ║  ║  ║
        # cr: 3/══╩═══╩═╡ 0x3 ╞═╩══╩══╩═
        #         1   0 └─────┘ 0  1  2
        circuit_if_false = QuantumCircuit(qr, cr)
        circuit_if_false.x(qr[0])
        circuit_if_false.measure(qr[0], cr[0])
        circuit_if_false.measure(qr[1], cr[1])
        with self.assertWarns(DeprecationWarning):
            circuit_if_false.x(qr[2]).c_if(cr, 0x3)
        circuit_if_false.measure(qr[0], cr[0])
        circuit_if_false.measure(qr[1], cr[1])
        circuit_if_false.measure(qr[2], cr[2])
        job = self.backend.run(
            transpile([circuit_if_true, circuit_if_false], self.backend),
            shots=shots,
            seed_simulator=self.seed,
        )
        result = job.result()
        counts_if_true = result.get_counts(circuit_if_true)
        counts_if_false = result.get_counts(circuit_if_false)
        self.assertEqual(counts_if_true, {"111": 100})
        self.assertEqual(counts_if_false, {"001": 100})

    def test_bit_cif_crossaffect(self):
        """Test if bits in a classical register other than
        the single conditional bit affect the conditioned operation."""
        #               ┌───┐          ┌─┐
        # q0_0: ────────┤ H ├──────────┤M├
        #       ┌───┐   └─╥─┘    ┌─┐   └╥┘
        # q0_1: ┤ X ├─────╫──────┤M├────╫─
        #       ├───┤     ║      └╥┘┌─┐ ║
        # q0_2: ┤ X ├─────╫───────╫─┤M├─╫─
        #       └───┘┌────╨─────┐ ║ └╥┘ ║
        # c0: 3/═════╡ c0_0=0x1 ╞═╩══╩══╬═
        #            └──────────┘ 1  2  ║
        # c1: 1/════════════════════════╩═
        #                               0
        shots = 100
        qr = QuantumRegister(3)
        cr = ClassicalRegister(3)
        cr1 = ClassicalRegister(1)
        circuit = QuantumCircuit(qr, cr, cr1)
        circuit.x([qr[1], qr[2]])
        circuit.measure(qr[1], cr[1])
        circuit.measure(qr[2], cr[2])
        with self.assertWarns(DeprecationWarning):
            circuit.h(qr[0]).c_if(cr[0], True)
        circuit.measure(qr[0], cr1[0])
        job = self.backend.run(circuit, shots=shots, seed_simulator=self.seed)
        result = job.result().get_counts()
        target = {"0 110": 100}
        self.assertEqual(result, target)

    def test_teleport(self):
        """Test teleportation as in tutorials"""
        #       ┌─────────┐          ┌───┐ ░ ┌─┐
        # qr_0: ┤ Ry(π/4) ├───────■──┤ H ├─░─┤M├────────────────────
        #       └──┬───┬──┘     ┌─┴─┐└───┘ ░ └╥┘┌─┐
        # qr_1: ───┤ H ├─────■──┤ X ├──────░──╫─┤M├─────────────────
        #          └───┘   ┌─┴─┐└───┘      ░  ║ └╥┘ ┌───┐  ┌───┐ ┌─┐
        # qr_2: ───────────┤ X ├───────────░──╫──╫──┤ Z ├──┤ X ├─┤M├
        #                  └───┘           ░  ║  ║  └─╥─┘  └─╥─┘ └╥┘
        #                                     ║  ║ ┌──╨──┐   ║    ║
        # cr0: 1/═════════════════════════════╩══╬═╡ 0x1 ╞═══╬════╬═
        #                                     0  ║ └─────┘┌──╨──┐ ║
        # cr1: 1/════════════════════════════════╩════════╡ 0x1 ╞═╬═
        #                                        0        └─────┘ ║
        # cr2: 1/═════════════════════════════════════════════════╩═
        #                                                         0
        self.log.info("test_teleport")
        pi = np.pi
        shots = 4000
        qr = QuantumRegister(3, "qr")
        cr0 = ClassicalRegister(1, "cr0")
        cr1 = ClassicalRegister(1, "cr1")
        cr2 = ClassicalRegister(1, "cr2")
        circuit = QuantumCircuit(qr, cr0, cr1, cr2, name="teleport")
        circuit.h(qr[1])
        circuit.cx(qr[1], qr[2])
        circuit.ry(pi / 4, qr[0])
        circuit.cx(qr[0], qr[1])
        circuit.h(qr[0])
        circuit.barrier(qr)
        circuit.measure(qr[0], cr0[0])
        circuit.measure(qr[1], cr1[0])
        with self.assertWarns(DeprecationWarning):
            circuit.z(qr[2]).c_if(cr0, 1)
        with self.assertWarns(DeprecationWarning):
            circuit.x(qr[2]).c_if(cr1, 1)
        circuit.measure(qr[2], cr2[0])
        job = self.backend.run(
            transpile(circuit, self.backend), shots=shots, seed_simulator=self.seed
        )
        results = job.result()
        data = results.get_counts("teleport")
        alice = {
            "00": data["0 0 0"] + data["1 0 0"],
            "01": data["0 1 0"] + data["1 1 0"],
            "10": data["0 0 1"] + data["1 0 1"],
            "11": data["0 1 1"] + data["1 1 1"],
        }
        bob = {
            "0": data["0 0 0"] + data["0 1 0"] + data["0 0 1"] + data["0 1 1"],
            "1": data["1 0 0"] + data["1 1 0"] + data["1 0 1"] + data["1 1 1"],
        }
        self.log.info("test_teleport: circuit:")
        self.log.info(dumps(circuit))
        self.log.info("test_teleport: data %s", data)
        self.log.info("test_teleport: alice %s", alice)
        self.log.info("test_teleport: bob %s", bob)
        alice_ratio = 1 / np.tan(pi / 8) ** 2
        bob_ratio = bob["0"] / float(bob["1"])
        error = abs(alice_ratio - bob_ratio) / alice_ratio
        self.log.info("test_teleport: relative error = %s", error)
        self.assertLess(error, 0.05)

    def test_memory(self):
        """Test memory."""
        #       ┌───┐        ┌─┐
        # qr_0: ┤ H ├──■─────┤M├───
        #       └───┘┌─┴─┐   └╥┘┌─┐
        # qr_1: ─────┤ X ├────╫─┤M├
        #            └┬─┬┘    ║ └╥┘
        # qr_2: ──────┤M├─────╫──╫─
        #       ┌───┐ └╥┘ ┌─┐ ║  ║
        # qr_3: ┤ X ├──╫──┤M├─╫──╫─
        #       └───┘  ║  └╥┘ ║  ║
        # cr0: 2/══════╬═══╬══╩══╩═
        #              ║   ║  0  1
        #              ║   ║
        # cr1: 2/══════╩═══╩═══════
        #              0   1
        qr = QuantumRegister(4, "qr")
        cr0 = ClassicalRegister(2, "cr0")
        cr1 = ClassicalRegister(2, "cr1")
        circ = QuantumCircuit(qr, cr0, cr1)
        circ.h(qr[0])
        circ.cx(qr[0], qr[1])
        circ.x(qr[3])
        circ.measure(qr[0], cr0[0])
        circ.measure(qr[1], cr0[1])
        circ.measure(qr[2], cr1[0])
        circ.measure(qr[3], cr1[1])

        shots = 50
        job = self.backend.run(
            transpile(circ, self.backend), shots=shots, seed_simulator=self.seed, memory=True
        )
        result = job.result()
        memory = result.get_memory()
        self.assertEqual(len(memory), shots)
        for mem in memory:
            self.assertIn(mem, ["10 00", "10 11"])

    def test_unitary(self):
        """Test unitary gate instruction"""
        max_qubits = 4
        x_mat = np.array([[0, 1], [1, 0]])
        # Test 1 to max_qubits for random n-qubit unitary gate
        for i in range(max_qubits):
            num_qubits = i + 1
            # Apply X gate to all qubits
            multi_x = x_mat
            for _ in range(i):
                multi_x = np.kron(multi_x, x_mat)
            # Target counts
            shots = 1024
            target_counts = {num_qubits * "1": shots}
            # Test circuit
            qr = QuantumRegister(num_qubits, "qr")
            cr = ClassicalRegister(num_qubits, "cr")
            circuit = QuantumCircuit(qr, cr)
            circuit.unitary(multi_x, qr)
            circuit.measure(qr, cr)
            job = self.backend.run(transpile(circuit, self.backend), shots=shots)
            result = job.result()
            counts = result.get_counts(0)
            self.assertEqual(counts, target_counts)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Tests for utility functions.
"""

import unittest
import random
import numpy as np

import qiskit.synthesis.unitary.aqc.fast_gradient.fast_grad_utils as myu
from qiskit.synthesis.unitary.aqc.elementary_operations import rx_matrix as _rx
from qiskit.synthesis.unitary.aqc.elementary_operations import ry_matrix as _ry
from qiskit.synthesis.unitary.aqc.elementary_operations import rz_matrix as _rz
from test import QiskitTestCase  # pylint: disable=wrong-import-order
import test.python.synthesis.aqc.fast_gradient.utils_for_testing as tut  # pylint: disable=wrong-import-order


class TestUtils(QiskitTestCase):
    """
    Tests utility functions used by the fast gradient implementation.
    """

    max_nqubits_rev_bits = 7  # max. number of qubits in test_reverse_bits
    max_nqubits_mk_unit = 5  # max. number of qubits in test_make_unit_vector
    max_nqubits_swap = 7  # max. number of qubits in test_swap
    max_nqubits_perms = 5  # max. number of qubits in test_permutations
    num_repeats_gates2x2 = 100  # number of repetitions in test_gates2x2

    def setUp(self):
        super().setUp()
        np.random.seed(0x0696969)

    def test_reverse_bits(self):
        """
        Tests the function utils.reverse_bits().
        """

        nbits = self.max_nqubits_rev_bits
        for x in range(2**nbits):
            y = myu.reverse_bits(x, nbits, enable=True)
            self.assertTrue(x == myu.reverse_bits(y, nbits, enable=True))
            for i in range(nbits):
                self.assertTrue(((x >> i) & 1) == ((y >> (nbits - i - 1)) & 1))

        x = np.arange(2**nbits, dtype=np.int64)
        y = myu.reverse_bits(x, nbits, enable=True)
        self.assertTrue(np.all(myu.reverse_bits(y, nbits, enable=True) == x))

    def test_make_unit_vector(self):
        """
        Tests the function utils_for_testing.make_unit_vector().
        """

        # Computes unit vector as Kronecker product of (1 0) or (0 1) sub-vectors.
        def _make_unit_vector_simple(num: int, nbits: int) -> np.ndarray:
            self.assertTrue(isinstance(num, int) and isinstance(nbits, int))
            self.assertTrue(0 <= num < 2**nbits)
            vec = 1
            for i in range(nbits):
                x = (num >> i) & 1
                vec = np.kron(vec, np.asarray([1 - x, x], dtype=np.int64))
            self.assertTrue(vec.dtype == np.int64 and vec.shape == (2**nbits,))
            return vec

        for n in range(1, self.max_nqubits_mk_unit + 1):
            for k in range(2**n):
                v1 = _make_unit_vector_simple(k, n)
                v2 = tut.make_unit_vector(k, n)
                self.assertTrue(np.all(v1 == v2))

    def test_swap(self):
        """
        Tests the function utils.swap_bits().
        """

        nbits = self.max_nqubits_swap
        for x in range(2**nbits):
            for a in range(nbits):
                for b in range(nbits):
                    y = myu.swap_bits(x, a, b)
                    self.assertTrue(x == myu.swap_bits(y, a, b))
                    self.assertTrue(((x >> a) & 1) == ((y >> b) & 1))
                    self.assertTrue(((x >> b) & 1) == ((y >> a) & 1))

    def test_permutations(self):
        """
        Tests various properties of permutations and their compositions.
        """

        def _perm2_mat(perm: np.ndarray) -> np.ndarray:
            """Creates permutation matrix from a permutation."""
            return np.eye(perm.size, dtype=np.int64)[perm]

        def _permutation_and_matrix(size: int) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
            """Creates direct and inverse permutations and their matrices."""
            perm = np.random.permutation(np.arange(size, dtype=np.int64))
            inv_perm = myu.inverse_permutation(perm)
            p_mat, q_mat = _perm2_mat(perm), _perm2_mat(inv_perm)
            return perm, inv_perm, p_mat, q_mat

        for n in range(1, self.max_nqubits_perms + 1):

            dim = 2**n
            for _ in range(100):
                perm1, inv_perm1, p1_mat, q1_mat = _permutation_and_matrix(dim)
                perm2, inv_perm2, p2_mat, q2_mat = _permutation_and_matrix(dim)

                iden = tut.eye_int(n)
                self.assertTrue(np.all(p1_mat == q1_mat.T) and np.all(p1_mat.T == q1_mat))
                self.assertTrue(np.all(p2_mat == q2_mat.T) and np.all(p2_mat.T == q2_mat))
                self.assertTrue(np.all(p1_mat @ q1_mat == iden) and np.all(q1_mat @ p1_mat == iden))
                self.assertTrue(np.all(p2_mat @ q2_mat == iden) and np.all(q2_mat @ p2_mat == iden))

                self.assertTrue(np.all(perm2[perm1] == np.take(perm2, perm1)))
                self.assertTrue(np.all(perm1[perm2] == np.take(perm1, perm2)))

                self.assertTrue(np.all(p1_mat @ p2_mat == _perm2_mat(perm2[perm1])))
                self.assertTrue(np.all(p2_mat @ p1_mat == _perm2_mat(perm1[perm2])))

                self.assertTrue(np.all(p1_mat @ q2_mat == _perm2_mat(inv_perm2[perm1])))
                self.assertTrue(np.all(q2_mat @ p1_mat == _perm2_mat(perm1[inv_perm2])))

                self.assertTrue(np.all(p2_mat @ q1_mat == _perm2_mat(inv_perm1[perm2])))
                self.assertTrue(np.all(q1_mat @ p2_mat == _perm2_mat(perm2[inv_perm1])))

                _, inv_perm3, p3_mat, _ = _permutation_and_matrix(dim)
                perm4, _, _, q4_mat = _permutation_and_matrix(dim)

                mat = tut.rand_matrix(dim=dim, kind="randint")

                pmq = p1_mat @ mat @ q2_mat
                self.assertTrue(np.all(pmq == mat[perm1][:, perm2]))
                self.assertTrue(np.all(pmq == np.take(np.take(mat, perm1, axis=0), perm2, axis=1)))
                pmq = p1_mat @ mat @ p2_mat
                self.assertTrue(np.all(pmq == mat[perm1][:, inv_perm2]))
                self.assertTrue(
                    np.all(pmq == np.take(np.take(mat, perm1, axis=0), inv_perm2, axis=1))
                )

                ppmpq = p1_mat @ p2_mat @ mat @ p3_mat @ q4_mat
                self.assertTrue(np.all(ppmpq == mat[perm2][perm1][:, inv_perm3][:, perm4]))
                self.assertTrue(np.all(ppmpq == mat[perm2[perm1]][:, inv_perm3[perm4]]))
                self.assertTrue(
                    np.all(
                        ppmpq
                        == np.take(np.take(mat, perm2[perm1], axis=0), inv_perm3[perm4], axis=1)
                    )
                )

    def test_gates2x2(self):
        """
        Tests implementation of 2x2 gate matrices vs the ones that create and
        return temporary array upon every call (but have more compact code).
        """

        tol = np.finfo(np.float64).eps * 2.0
        out = np.full((2, 2), fill_value=0, dtype=np.complex128)
        for test in range(self.num_repeats_gates2x2):  # pylint: disable=unused-variable
            phi = random.random() * 2.0 * np.pi
            self.assertTrue(np.allclose(myu.make_rx(phi, out=out), _rx(phi), atol=tol, rtol=tol))
            self.assertTrue(np.allclose(myu.make_ry(phi, out=out), _ry(phi), atol=tol, rtol=tol))
            self.assertTrue(np.allclose(myu.make_rz(phi, out=out), _rz(phi), atol=tol, rtol=tol))


if __name__ == "__main__":
    unittest.main()
